"""Module for formulating a :external+sklearn:py:class:`sklearn.preprocessing.StandardScalar`,
:external+sklearn:py:class:`sklearn.preprocessing.PolynomialFeatures`,
:external+sklearn:py:class:`sklearn.preprocessing.Normalizer`, or
:external+sklearn:py:class:`sklearn.preprocessing.Binarizer`
into a PySCIPOpt Model.
"""

import numpy as np
from pyscipopt import sqrt

from ..exceptions import NoModel
from ..modelling.classification import max_formulation
from .skgetter import SKtransformer


def add_binarizer_constr(
    scip_model,
    binarizer,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate binarizer into scip_model.

    Note that this function creates the output variables from
    the input variables.

    Parameters
    ----------
    scip_model : SCIP Model
        The PySCIPOpt Model where the polynomial features should be inserted.
    binarizer : :external+sklearn:py:class:`sklearn.preprocessing.Binarizer`
        The normalizer to insert into scip_model.
    input_vars : list or np.ndarray
        Decision variables used as input for the transformer in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for the transformer in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    sklearn.preprocessing.BinarizerConstr
        Object containing information about what was added to scip_model to insert the
        binarizer in it
    """
    return BinarizerConstr(
        scip_model, binarizer, input_vars, output_vars, unique_naming_prefix, **kwargs
    )


def add_normalizer_constr(
    scip_model,
    normalizer,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate normalizer into scip_model.

    Note that this function creates the output variables from
    the input variables.

    Parameters
    ----------
    scip_model : SCIP Model
        The PySCIPOpt Model where the polynomial features should be inserted.
    normalizer : :external+sklearn:py:class:`sklearn.preprocessing.Normalizer`
        The normalizer to insert into scip_model.
    input_vars : list or np.ndarray
        Decision variables used as input for the transformer in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for the transformer in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    sklearn.preprocessing.NormalizerConstr
        Object containing information about what was added to scip_model to insert the
        normalizer in it
    """
    return NormalizerConstr(
        scip_model, normalizer, input_vars, output_vars, unique_naming_prefix, **kwargs
    )


def add_polynomial_features_constr(
    scip_model,
    polynomial_features,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate polynomial_features into scip_model.

    Note that this function creates the output variables from
    the input variables.

    Parameters
    ----------
    scip_model : SCIP Model
        The PySCIPOpt Model where the polynomial features should be inserted.
    polynomial_features : :external+sklearn:py:class:`sklearn.preprocessing.PolynomialFeatures`
        The polynomial features to insert into scip_model.
    input_vars : list or np.ndarray
        Decision variables used as input for the transformer in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for the transformer in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    sklearn.preprocessing.PolynomialFeaturesConstr
        Object containing information about what was added to scip_model to insert the
        polynomial_features in it
    """
    return PolynomialFeaturesConstr(
        scip_model, polynomial_features, input_vars, output_vars, unique_naming_prefix, **kwargs
    )


def add_standard_scaler_constr(
    scip_model, standard_scaler, input_vars, output_vars=None, unique_naming_prefix="", **kwargs
):
    """Formulate standard_scaler into scip_model.

    Note that this function creates the output variables from
    the input variables.

    Parameters
    ----------
    scip_model : SCIP Model
        The PySCIPOpt Model where the polynomial features should be inserted.
    standard_scaler : :external+sklearn:py:class:`sklearn.preprocessing.StandardScaler`
        The standard scalar to insert into scip_model.
    input_vars : list or np.ndarray
        Decision variables used as input for the transformer in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for the transformer in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    sklearn.preprocessing.StandardScalerConstr
        Object containing information about what was added to scip_model to insert the
        standard_scaler in it
    """
    return StandardScalerConstr(
        scip_model, standard_scaler, input_vars, output_vars, unique_naming_prefix, **kwargs
    )


class StandardScalerConstr(SKtransformer):
    """Class to formulate a fitted
    :external+sklearn:py:class:`sklearn.preprocessing.StandardScaler` in a
    PySCIPOpt Model.
    """

    def __init__(
        self, scip_model, scaler, input_vars, output_vars, unique_naming_prefix, **kwargs
    ):
        self.output_size = scaler.n_features_in_
        super().__init__(
            scip_model, scaler, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Do the transformation on x."""

        scale = self.transformer.scale_
        mean = self.transformer.mean_

        scaler_cons = np.zeros(self.input.shape, dtype=object)

        for i in range(self.input.shape[0]):
            for j in range(self.input.shape[-1]):
                name = self.unique_naming_prefix + f"_scaler_{i}_{j}"
                scaler_cons[i][j] = self.scip_model.addCons(
                    self.input[i][j] - (scale[j] * self.output[i][j]) == mean[j], name=name
                )

        self._created_cons.append(scaler_cons)


class PolynomialFeaturesConstr(SKtransformer):
    """Class to formulate a fitted
    :external+sklearn:py:class:`sklearn.preprocessing.PolynomialFeatures` in a
    PySCIPOpt Model.
    """

    def __init__(
        self,
        scip_model,
        polynomial_features,
        input_vars,
        output_vars,
        unique_naming_prefix,
        **kwargs,
    ):
        self.output_size = polynomial_features.n_output_features_
        super().__init__(
            scip_model,
            polynomial_features,
            input_vars,
            output_vars,
            unique_naming_prefix,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        """Do the transformation on x."""

        n_samples, n_features = self.input.shape
        powers = self.transformer.powers_
        assert powers.shape[0] == self.transformer.n_output_features_
        assert powers.shape[1] == n_features

        poly_cons = np.zeros((n_samples, self.transformer.n_output_features_), dtype=object)

        for i in range(n_samples):
            for j, power in enumerate(powers):
                scip_expr = 1.0
                for k, input_var in enumerate(self.input[i, :]):
                    if power[k] >= 1:
                        scip_expr *= input_var ** power[k]
                name = self.unique_naming_prefix + f"_poly_{i}_{j}"
                poly_cons[i][j] = self.scip_model.addCons(
                    self.output[i][j] == scip_expr, name=name
                )

        self._created_cons.append(poly_cons)


class NormalizerConstr(SKtransformer):
    """Class to formulate a fitted
    :external+sklearn:py:class:`sklearn.preprocessing.Normalizer` in a
    PySCIPOpt Model.
    """

    def __init__(
        self, scip_model, normalizer, input_vars, output_vars, unique_naming_prefix, **kwargs
    ):
        self.output_size = normalizer.n_features_in_
        if normalizer.norm not in ["l1", "l2", "max"]:
            raise NoModel(normalizer, f"Norm {normalizer.norm} is not supported.")
        super().__init__(
            scip_model, normalizer, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Do the transformation on x."""

        norm = self.transformer.norm
        n_samples = self.input.shape[0]
        n_features = self.input.shape[-1]

        # Determine if slack variables need to be created
        pos_vars = True
        if norm in ["l1", "max"]:
            for i in range(n_samples):
                for j in range(n_features):
                    if self.input[i][j].getLbGlobal() < 0:
                        pos_vars = False
                        break
        if not pos_vars:
            pos_slack_vars = self.scip_model.addMatrixVar(
                self.input.shape, vtype="C", lb=0, ub=None, name=self.unique_naming_prefix + "_pos"
            )
            neg_slack_vars = self.scip_model.addMatrixVar(
                self.input.shape, vtype="C", lb=0, ub=None, name=self.unique_naming_prefix + "_neg"
            )
            dist_sum_cons = self.scip_model.addMatrixCons(
                self.input == pos_slack_vars - neg_slack_vars,
                name=self.unique_naming_prefix + "_dist_slack",
            )
            self._created_vars.append(pos_slack_vars)
            self._created_vars.append(neg_slack_vars)
            self._created_cons.append(dist_sum_cons)
        # If using max norm then need to get the max
        if norm == "max":
            max_scale_vars = self.scip_model.addMatrixVar(
                (n_samples, 1),
                vtype="C",
                lb=None,
                ub=None,
                name=self.unique_naming_prefix + "_max",
            )
            input_vars = self.input if pos_vars else pos_slack_vars + neg_slack_vars
            max_form_vars, max_form_cons = max_formulation(
                self.scip_model,
                input_vars,
                max_scale_vars,
                self.unique_naming_prefix,
            )
            for scip_var in max_form_vars:
                self._created_vars.append(scip_var)
            for scip_cons in max_form_cons:
                self._created_cons.append(scip_cons)

        # Create the normalisation constraints
        norm_cons = np.zeros(self.input.shape, dtype=object)
        for i in range(n_samples):
            if norm == "l2":
                scale = sqrt(sum(self.input[i][j] ** 2 for j in range(n_features)))
            elif norm == "l1":
                if pos_vars:
                    scale = sum(self.input[i][j] for j in range(n_features))
                else:
                    scale = sum(
                        pos_slack_vars[i][j] + neg_slack_vars[i][j] for j in range(n_features)
                    )
            elif norm == "max":
                scale = max_scale_vars[i][0]
            else:
                raise NoModel(self.transformer, f"Norm {norm} is not supported.")
            for j in range(n_features):
                name = f"norm_cons_{i}_{j}"
                norm_cons[i][j] = self.scip_model.addCons(
                    self.output[i][j] == self.input[i][j] / scale, name=name
                )

        self._created_cons.append(norm_cons)


class BinarizerConstr(SKtransformer):
    """Class to formulate a fitted
    :external+sklearn:py:class:`sklearn.preprocessing.Binarizer` in a
    PySCIPOpt Model.
    """

    def __init__(
        self, scip_model, binarizer, input_vars, output_vars, unique_naming_prefix, **kwargs
    ):
        self.output_size = binarizer.n_features_in_
        self.classification = True
        super().__init__(
            scip_model, binarizer, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Do the transformation on x."""

        n_samples = self.input.shape[0]
        n_features = self.input.shape[-1]
        threshold = self.transformer.threshold

        binary_indicator_cons = np.zeros((n_samples, n_features, 2), dtype=object)

        for i in range(self.input.shape[0]):
            for j in range(self.input.shape[-1]):
                name = self.unique_naming_prefix + f"_binary_ind_{i}_{j}_0"
                binary_indicator_cons[i][j][0] = self.scip_model.addConsIndicator(
                    self.input[i][j] <= threshold, self.output[i][j], False, name=name
                )
                name = self.unique_naming_prefix + f"_binary_ind_{i}_{j}_1"
                binary_indicator_cons[i][j][1] = self.scip_model.addConsIndicator(
                    -self.input[i][j] <= -threshold, self.output[i][j], True, name=name
                )

        self._created_cons.append(binary_indicator_cons)


def sklearn_transformers():
    """Return dictionary of Scikit Learn preprocessing objects."""
    return {
        "StandardScaler": add_standard_scaler_constr,
        "PolynomialFeatures": add_polynomial_features_constr,
        "Normalizer": add_normalizer_constr,
        "Binarizer": add_binarizer_constr,
    }



import numpy as np
import pytest
from pyscipopt import Model
from sklearn.cross_decomposition import PLSCanonical, PLSRegression
from sklearn.datasets import (
    load_breast_cancer,
    load_diabetes,
    load_iris,
    make_regression,
)
from sklearn.ensemble import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import (
    ElasticNet,
    Lasso,
    LinearRegression,
    LogisticRegression,
    Ridge,
)
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import LinearSVC, LinearSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from src.pyscipopt_ml import add_predictor_constr

"""
Tests for the various possible input types of each ML Model from sklearn.
load_iris is used for multi-classification, load_diabetes used for regression,
load_breast_cancer used for single class classification, and laod_wine is used for
multiple output regression.

For all models tests are run on input types being lists and numpy arrays of various dimensions.
Tests are additionally run with and without the output defined by the user.
"""


def get_data_and_create_basic_model(
    predictor,
    multi_dimension=False,
    classification=False,
    n_samples=1,
    input_type="numpy",
    output_type="numpy",
):
    assert input_type in ("numpy", "matrix", "list")
    assert output_type in ("numpy", "matrix", "list", "None")

    # Load the correct data
    if multi_dimension and classification:
        X, y = load_iris(return_X_y=True)
        input_size = 4
        output_size = 3
    elif multi_dimension and not classification:
        X, y = make_regression(
            n_samples=500, n_features=10, n_informative=5, n_targets=2, random_state=0, noise=0.5
        )
        input_size = 10
        output_size = 2
    elif not multi_dimension and classification:
        X, y = load_breast_cancer(return_X_y=True)
        input_size = 30
        output_size = 1
    else:
        X, y = load_diabetes(return_X_y=True)
        input_size = 10
        output_size = 1

    # Fit the predictor
    predictor = predictor().fit(X, y)

    # Create the SCIP model
    scip = Model()

    # Create the input variables
    if input_type in ["numpy", "list"]:
        if input_type == "numpy":
            input_vars = np.zeros(shape=(n_samples, input_size), dtype=object)
        else:
            input_vars = [[None for _ in range(input_size)] for _ in range(n_samples)]
        for i in range(n_samples):
            for j in range(input_size):
                input_vars[i][j] = scip.addVar(name=f"x_{i}", vtype="C", lb=-10, ub=10)
    else:
        input_vars = scip.addMatrixVar((n_samples, input_size), name="x", vtype="C", lb=-10, ub=10)

    # Create the output variables
    if output_type in ["numpy", "list", "None"]:
        if output_type == "numpy":
            output_vars = np.zeros(shape=(n_samples, output_size), dtype=object)
        elif output_type == "list":
            output_vars = [[None for _ in range(output_size)] for _ in range(n_samples)]
        else:
            output_vars = None
        if output_type != "None":
            for i in range(n_samples):
                for j in range(output_size):
                    output_vars[i][j] = scip.addVar(name=f"y_{i}", vtype="C", lb=-10, ub=10)
    else:
        output_vars = scip.addMatrixVar(
            (n_samples, output_size), name="y", vtype="C", lb=-10, ub=10
        )

    pred_cons = add_predictor_constr(scip, predictor, input_vars, output_vars)


testdata = [
    (LinearRegression, True, False, 2, "numpy", "numpy"),
    (LinearRegression, True, False, 2, "matrix", "matrix"),
    (ElasticNet, False, False, 1, "numpy", "numpy"),
    (LinearRegression, False, False, 2, "numpy", "numpy"),
    (Lasso, True, False, 1, "list", "numpy"),
    (LinearRegression, True, False, 2, "numpy", "list"),
    (Ridge, True, False, 2, "numpy", "None"),
    (LogisticRegression, True, True, 2, "numpy", "numpy"),
    (LogisticRegression, False, True, 1, "numpy", "numpy"),
    (LogisticRegression, False, True, 1, "numpy", "matrix"),
    (LinearRegression, True, False, 2, "list", "None"),
    (PLSRegression, True, False, 1, "numpy", "numpy"),
    (PLSCanonical, True, False, 2, "numpy", "None"),
    (DecisionTreeRegressor, True, False, 2, "numpy", "numpy"),
    (DecisionTreeRegressor, False, False, 1, "numpy", "numpy"),
    (DecisionTreeRegressor, False, False, 1, "matrix", "None"),
    (DecisionTreeRegressor, True, False, 2, "list", "None"),
    (DecisionTreeClassifier, True, True, 2, "numpy", "numpy"),
    (DecisionTreeClassifier, False, True, 2, "list", "None"),
    (GradientBoostingRegressor, False, False, 2, "numpy", "numpy"),
    (GradientBoostingRegressor, False, False, 2, "list", "None"),
    (GradientBoostingClassifier, True, True, 2, "numpy", "numpy"),
    (GradientBoostingClassifier, False, True, 2, "list", "None"),
    (RandomForestRegressor, True, False, 2, "numpy", "numpy"),
    (RandomForestRegressor, False, False, 1, "numpy", "numpy"),
    (RandomForestRegressor, False, False, 2, "list", "None"),
    (RandomForestClassifier, False, True, 1, "numpy", "numpy"),
    (RandomForestClassifier, False, True, 1, "list", "matrix"),
    (RandomForestClassifier, False, True, 2, "list", "None"),
    (MLPRegressor, True, False, 1, "numpy", "numpy"),
    (MLPRegressor, False, False, 2, "list", "None"),
    (MLPClassifier, True, True, 1, "numpy", "numpy"),
    (MLPClassifier, False, True, 2, "list", "None"),
    (LinearSVR, False, False, 2, "list", "None"),
    (LinearSVC, False, True, 1, "numpy", "numpy"),
]


@pytest.mark.parametrize(
    "predictor,multi_dimension,classification,n_samples,input_type,output_type", testdata
)
def test_input_output(
    predictor, multi_dimension, classification, n_samples, input_type, output_type
):
    get_data_and_create_basic_model(
        predictor, multi_dimension, classification, n_samples, input_type, output_type
    )



"""Module for formulating :external+onnx:py:class:`onnx.ModelProto` model in a PySCIPOPT Model.

Thanks to https://github.com/cog-imperial/OMLT/blob/main/src/omlt/io/onnx_parser.py for help with the parser

Notes to help with reading:
- This code is more general than necessary. It has the general structure (still missing parts) to later allow not
just fully connected layers. Many assert statements would need to be removed and code would need to be thoroughly
checked before introducing such features.
- A "node" here is used in the ONNX content, and can represent an entire layer.
"""

import numpy as np
import onnxruntime as ort
from onnx import numpy_helper

from ..exceptions import NoModel, NoSolution, ParameterError
from ..modelling.classification import argmax_bound_formulation
from ..modelling.neuralnet import AbstractNNLayer, BaseNNConstr


def add_onnx_constr(
    scip_model,
    onnx_model,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    output_type="regression",
    **kwargs,
):
    """Formulate an onnx ModelProto into scip_model.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the ONNX ModelProto should be inserted.

    onnx_model : :external+onnx:py:class:`onnx.ModelProto`
        The onnx model to insert as predictor.

    input_vars : np.ndarray
        Decision variables used as input for the feed forward neural network in PySCIPOpt Model.

    output_vars : np.ndarray
        Decision variables used as output for the feed forward neural network in the PySCIPOpt Model

    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    output_type : {"classification", "regression"}, default="regression"
        If the option chosen is "classification" the output is 1 for exactly one class and
        0 for all others. If the option chosen is "regression" then the output for
        each node of the final layer is the value from the ONNX model.

    Returns
    -------
    ONNXConstr
        Object containing information about what was added to model to insert the
        predictor into it

    Raises
    ------
    NoModel
        If a section of the ONNX model structure (layer or activation) is not implemented.

    Warning
    -------
    Only MatMul, Gemm, Relu, Sigmoid, Tanh, Softmax, and Softplus operator types are supported.
    Add, Cast, and Reshape are also often possible, although not in arbitrary contexts.

    Note
    ----
    |VariablesDimensionsWarn|

    """
    return ONNXConstr(
        scip_model,
        onnx_model,
        input_vars,
        output_vars,
        unique_naming_prefix,
        output_type,
        **kwargs,
    )


class ONNXConstr(BaseNNConstr):
    """
    Transform an ONNX ModelProto that represents a fully connected feed forward neural network
     to SCIP constraints with input and output as matrices of variables.

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        output_type="regression",
        **kwargs,
    ):
        if output_type not in ["regression", "classification"]:
            raise NoModel(
                predictor, f"Output type {output_type} is neither regression nor classification"
            )
        self.output_type = output_type
        self._graph = predictor.graph

        # Check that the graph from ONNX contains only nodes that we can work with
        for node in self._graph.node:
            if node.op_type == "MatMul":
                if len(node.input) != 2:
                    raise ValueError(
                        f"MatMul node {node.name} input has {len(node.input)} many dimensions, not 2."
                    )
            elif node.op_type == "Gemm":
                if len(node.input) != 3:
                    raise ValueError(
                        f"Gemm node {node.name} input has {len(node.input)} many dimensions, not 3."
                    )
            elif node.op_type not in [
                "Add",
                "Relu",
                "Sigmoid",
                "Tanh",
                "Softmax",
                "Softplus",
                "Reshape",
                "Cast",
            ]:
                raise NoModel(predictor, f"ONNX node op type {node.op_type} not supported")

        # Verify the input and output dimensions
        if len(self._graph.input) != 1:
            raise NoModel(
                predictor,
                f"There are {len(self._graph.input)} many input tensors. Currently only one input vector"
                f" is supported",
            )
        if len(self._graph.output) != 1:
            raise NoModel(
                predictor,
                f"There are {len(self._graph.output)} many output tensors. Currently only one output vector"
                f" is supported",
            )

        # The initializers contain all constant data, e.g., weights and biases (is often input to a node)
        self._initializers = {}
        for initializer in self._graph.initializer:
            self._initializers[initializer.name] = numpy_helper.to_array(initializer)

        # Derive the input size of the model. Do this by flattening along all dimensions
        # Simultaneously populate the nodes and node_by_output dictionaries
        # The nodes dictionary contains tuples "input" or "node", ONNX node, list of nodes it feeds into)
        # The nodes_by_output dictionary contains a mapping from an output name to a node name
        # An output in this context is likely going to be represented by a set of created variables
        # Note: In the case of the node being an original input node the output of the node maps to itself
        self.input_size = None
        self._nodes_by_output = {}
        self._nodes = {}
        for onnx_input in self._graph.input:
            for dim in onnx_input.type.tensor_type.shape.dim:
                dim_value = dim.dim_value
                if dim_value > 0:
                    self._nodes_by_output[onnx_input.name] = onnx_input.name
                    self._nodes[onnx_input.name] = ("input", onnx_input.type, [])
                    if self.input_size is None:
                        self.input_size = dim_value
                    else:
                        self.input_size *= dim_value
        # Further populate the nodes and nodes_by_output dictionaries
        for node in self._graph.node:
            for output in node.output:
                self._nodes_by_output[output] = node.name
        for node in self._graph.node:
            # Add node not connected to anything
            self._nodes[node.name] = ("node", node, [])

            # Map inputs by their output name
            node_inputs = [
                self._nodes_by_output[onnx_input]
                for onnx_input in node.input
                if onnx_input not in self._initializers
            ]

            if len(node_inputs) > 0:
                # Now connect inputs to the current node
                for onnx_input in node_inputs:
                    self._nodes[onnx_input][2].append(node.name)

        # Determine the output size
        # This is non-trivial..... It cannot be taken from the output of the graph
        # We therefore have to work backwards and find the final matrix multiplication or addition
        self.output_size = None
        node = self._graph.node[-1]
        max_depth = len(self._graph.node)
        i = 0
        while i <= max_depth:
            # If the node has no dimensional information then find a previous node that feeds into it
            if node.op_type not in ["Add", "MatMul", "Gemm"]:
                if node.name not in self._nodes_by_output:
                    for node_input_name in node.input:
                        if node_input_name in self._nodes_by_output:
                            node = self._nodes[self._nodes_by_output[node_input_name]][1]
                else:
                    node = self._nodes[self._nodes_by_output[node.name]][1]
                i += 1
                continue
            if node.op_type == "Add":
                for onnx_input in node.input:
                    if onnx_input in self._initializers:
                        self.output_size = self._initializers[onnx_input].shape[-1]
                        break
                if self.output_size is not None:
                    break
            elif node.op_type == "MatMul":
                [_, matrix_name] = list(node.input)
                self.output_size = self._initializers[matrix_name].shape[-1]
                break
            elif node.op_type == "Gemm":
                attr = _collect_attributes(node)
                [_, _, bias_name] = list(node.input)
                self.output_size = self._initializers[bias_name].shape[-1]
                break
            i += 1
            node = self._nodes[self._nodes_by_output[node.name]][1]
        if self.output_size is None:
            self.output_size = self.input_size

        super().__init__(
            scip_model, predictor, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Add the predictor constraints to SCIP."""

        # Initialise the node input to layer map. This is used to get the input variables for later layers
        self._node_layer_map = {}
        for onnx_input in self._graph.input:
            self._node_layer_map[onnx_input.name] = self

        # Traverse the graph. This begins by populating the node stack with the input nodes
        input_nodes = set()
        for onnx_input in self._graph.input:
            input_nodes.add(onnx_input.name)
        self._node_stack = list(input_nodes)
        # Keep track of the number of processed nodes for naming purposes
        self.num_processed_nodes = 0
        # Keep visiting a new node until the stack has been completely explored. As the ONNX graph is directed and
        # acyclic this will always terminate
        while len(self._node_stack) > 0:
            node_name = self._node_stack.pop()
            type_, node, next_nodes = self._nodes[node_name]
            # Skip cases where node is input
            if type_ == "node":
                layer = self._visit_node(node, next_nodes)
            else:
                for node in next_nodes:
                    self._node_stack.append(node)

        # In the case of classification force the output to be binary with the argmax formulation
        n_samples = self.input.shape[0]
        if len(self._layers) > 0:
            input_vars = self._layers[-1].output
            activation = self._layers[-1].activation
        else:
            input_vars = self.input
            activation = "identity"
        if self.output_type == "classification":
            if activation == "sigmoid":
                one_dim_center = 0
            else:
                one_dim_center = 0.5
            new_vars, new_cons = argmax_bound_formulation(
                self.scip_model,
                input_vars,
                self.output,
                self.unique_naming_prefix,
                one_dim_center=one_dim_center,
            )
            for var_set in new_vars:
                self._created_vars.append(var_set)
            for cons_set in new_cons:
                self._created_cons.append(cons_set)
        else:
            # In the case of regression we need to tie the global output to the output of the final layer
            if input_vars.shape[-1] != self.output_size:
                raise ValueError(
                    f"Dimension of final layer != dimension of output vars. "
                    f"{input_vars.shape[-1]} != {self.output_size}"
                )
            match_output_cons = np.zeros(self.output.shape, dtype=object)
            for i in range(n_samples):
                for j in range(input_vars.shape[-1]):
                    name = self.unique_naming_prefix + f"match_out_{i}_{j}"
                    match_output_cons[i][j] = self.scip_model.addCons(
                        input_vars[i][j] == self.output[i][j], name=name
                    )
            self._created_cons.append(match_output_cons)

    def _visit_node(self, node, next_nodes):
        """
        Function for exploring / processing / parsing / consuming the current node

        Parameters
        ----------
        node : onnx.NodeProto
            The node that is currently being explored
        next_nodes : list
            The nodes next in line to be explored (output nodes of the current node)

        Returns
        -------
        layer : AbstractNNLayer or None
            The layer of the NN generated by the current node

        """
        # Separate the handling of node types by case
        if node.op_type == "MatMul":
            next_nodes, layer = self._consume_dense_nodes(node, next_nodes)
        elif node.op_type == "Gemm":
            next_nodes, layer = self._consume_gemm_dense_nodes(node, next_nodes)
        elif node.op_type == "Reshape" or node.op_type == "Cast":
            for node_input in node.input:
                if node_input not in self._initializers:
                    for node_output in node.output:
                        self._node_layer_map[node_output] = self._node_layer_map[node_input]
            layer = None
        else:
            raise Exception(f"Unhandled node type {node.op_type}")

        for node in next_nodes:
            self._node_stack.append(node)

        return layer

    def _consume_dense_nodes(self, node, next_nodes):
        """Starting from a MatMul node parse nodes to form a dense Ax + b node."""
        if node.op_type != "MatMul":
            raise ValueError(f"{node.name} is op type {node.op_type}, not MatMul.")

        # Extract input to the node
        [name_of_node_input_for_layer, matrix_name] = list(node.input)
        node_weights = self._initializers[matrix_name]

        if len(next_nodes) != 1:
            raise ValueError(
                f"Next nodes must have length 1, {next_nodes} has length {len(next_nodes)}"
            )

        # Expect 'Add' node ahead
        type_, node, next_nodes = self._nodes[next_nodes[0]]
        if type_ != "node":
            raise TypeError(f"Expected a node next, got a {type_} instead.")
        if node.op_type != "Add":
            raise ValueError(
                f"The first node after the MatMul node, {node.name}, is a {node.op_type} node, not Add."
            )

        # Extract biases
        [bias_name_1, bias_name_2] = list(node.input)
        if bias_name_1 in self._initializers:
            node_biases = self._initializers[bias_name_1]
        elif bias_name_2 in self._initializers:
            node_biases = self._initializers[bias_name_2]
        else:
            raise ValueError(f"Node inputs were not found in graph initializers.")

        if len(node_weights.shape) != 2:
            raise ValueError(f"Node weights must be a 2-dimensional matrix.")
        if node_weights.shape[1] != node_biases.shape[1]:
            raise ValueError(
                f"Dimension mismatch between weights {node_weights.shape} and biases {node_biases.shape}"
            )
        if len(node.output) != 1:
            raise ValueError(
                f"Node output is {node.output} but should consist of a single ONNX node."
            )

        # Now model the actual layer
        next_nodes, layer = self._add_dense_layer_matmul_gemm(
            self._node_layer_map[name_of_node_input_for_layer],
            node_weights,
            node_biases,
            node,
            node.name,
            next_nodes,
        )

        return next_nodes, layer

    def _consume_gemm_dense_nodes(self, node, next_nodes):
        """Starting from a Gemm node parse nodes to form a dense aAB + bC node."""
        if node.op_type != "Gemm":
            raise ValueError(f"{node.name} is op type {node.op_type}, not Gemm.")

        # Extract input to the node
        attr = _collect_attributes(node)
        if "alpha" in attr:
            alpha = attr["alpha"]
        else:
            alpha = 1.0
        if "beta" in attr:
            beta = attr["beta"]
        else:
            beta = 1.0
        [name_of_node_input_for_layer, matrix_name, bias_name] = list(node.input)
        node_weights = self._initializers[matrix_name]
        # Transpose B
        if "transB" in attr and attr["transB"] == 1:
            node_weights = np.transpose(node_weights)
        node_biases = self._initializers[bias_name]
        # Apply the appropriate scalars
        node_weights = node_weights * alpha
        node_biases = beta * node_biases

        # Now model the actual layer
        next_nodes, layer = self._add_dense_layer_matmul_gemm(
            self._node_layer_map[name_of_node_input_for_layer],
            node_weights,
            node_biases,
            node,
            name_of_node_input_for_layer,
            next_nodes,
        )

        return next_nodes, layer

    def _add_dense_layer_matmul_gemm(
        self, input_layer, weights, biases, node, node_name, next_nodes
    ):
        """
        Add a dense layer to the SCIP model using information extracted from a matmul or gemm node

        Parameters
        ----------
        input_layer : ONNXConstr or AbstractNNLayer
            The input layer whose output will input to the now created layer
        weights : np.ndarray
            The weights (coefficients) of the layer
        biases : np.ndarray
            The biases of the layer
        node : onnx.NodeProto
            The current node being considered (can be different from node_name due to matmul -> add etc)
        node_name : str
            The name of the node the layer is being constructed from
        next_nodes : list
            The list of nodes that should be explored after the current node is processed

        Returns
        -------
        next_nodes : list
            The list of nodes that should be explored after the current node is processed
        layer : AbstractNNLayer
            The created dense layer

        """

        # Determine the activation function of the layer (if one exists)
        activation = "identity"
        if isinstance(input_layer, ONNXConstr):
            input_vars = input_layer.input
        elif isinstance(input_layer, AbstractNNLayer):
            input_vars = input_layer.output
        else:
            raise ValueError(
                f"Input layer {input_layer} representing {node_name} is incorrect type"
            )
        if len(next_nodes) == 1:
            # check for appropriate activation function
            type_, maybe_node, maybe_next_nodes = self._nodes[next_nodes[0]]
            if maybe_node.op_type in ["Relu", "Sigmoid", "Tanh", "Softmax", "Softplus"]:
                node = maybe_node
                activation = maybe_node.op_type.lower()
                next_nodes = maybe_next_nodes

        output = self.scip_model.addMatrixVar(
            (input_vars.shape[0], weights.shape[-1]),
            vtype="C",
            lb=None,
            name=self.unique_naming_prefix + f"layer_{self.num_processed_nodes}_{node_name}",
        )
        self._created_vars.append(output)
        layer = self.add_dense_layer(
            input_vars,
            weights,
            biases,
            activation,
            output,
            unique_naming_prefix=self.unique_naming_prefix
            + f"linear_{self.num_processed_nodes}_{node_name}_",
        )

        # Increment the number of processed nodes and populate the dictionary with the newly created layer
        for node_output in node.output:
            self._node_layer_map[node_output] = layer
        self.num_processed_nodes += 1

        return next_nodes, layer

    def get_error(self, eps=None):
        """
        Returns error in SCIP's solution with respect to the actual output of the ONNX neural network

        Parameters
        ----------
        eps : float or int or None, optional
            The maximum allowed tolerance for a mismatch between the actual predictive model and SCIP.
            If the error is larger than eps an appropriate warning is printed

        Returns
        -------
        error: np.ndarray
            The absolute values of the difference between SCIP's solution and the trained ML model's output given
            the input as defined by SCIP. The matrix is the same dimension as the output of the trained predictor.
            Using ONNX / pyscipopt, the absolute difference between onnxruntime.InferenceSession(input)
            and scip.getVal(output).

        Raises
        ------
        NoSolution
            If SCIP has no solution (either was not optimized or is infeasible).
        """
        if self._has_solution:
            reg = ort.InferenceSession(
                self.predictor.SerializeToString(), providers=["CPUExecutionProvider"]
            )
            # Depending on the ONNX model double or float is expected. Numpy defaults to double.
            t_in = self.input_values
            if reg.get_inputs()[0].type == "tensor(float)":
                t_in = t_in.astype(np.float32)
            # ONNX ModelProto also comes with a batch size. Work around this.
            n_samples_expected = reg.get_inputs()[0].shape[0]
            n_samples = self.input.shape[0]
            if n_samples_expected is None or n_samples_expected == n_samples:
                t_out = reg.run(None, {self.predictor.graph.input[0].name: t_in})
                if isinstance(t_out, list):
                    t_out = t_out[0]
            elif n_samples_expected == 1:
                first_pass = True
                for i in range(n_samples):
                    t_out_temp = reg.run(
                        None, {self.predictor.graph.input[0].name: np.array([t_in[i]])}
                    )
                    if isinstance(t_out_temp, list):
                        t_out_temp = t_out_temp[0]
                    if first_pass:
                        t_out = t_out_temp
                    else:
                        t_out = np.vstack((t_out, t_out_temp))
                    first_pass = False
            else:
                raise ParameterError(
                    f"ONNX GraphProto has input dim {n_samples_expected}. It must be None, 1, or the "
                    f"same as the input {n_samples}"
                )
            # In the case of single large batches the output might all be concatenated. Reshape it.
            if t_out.shape != self.output_values.shape:
                t_out = t_out.reshape(self.output_values.shape)
            if self.output_type == "classification":
                argmax_indices = np.argmax(t_out, axis=1)
                t_out = np.zeros_like(t_out, dtype=int)
                t_out[np.arange(t_out.shape[0]), argmax_indices] = 1
            r_val = np.abs(t_out - self.output_values)
            if eps is not None and np.max(r_val) > eps:
                print(f"{t_out} != {self.output_values}")
            return r_val
        raise NoSolution()


def _collect_attributes(node):
    """
    Returns the correct values of the attributes stored for the specific node

    Parameters
    ----------
    node :
        The node object from the graph structure of ONNX.


    Returns
    -------
    value_dict : dict
        A dictionary containing the attribute values of the node

    """
    value_dict = {}
    for attr in node.attribute:
        if attr.type == 1:  # FLOAT
            value_dict[attr.name] = attr.f
        elif attr.type == 2:  # INT
            value_dict[attr.name] = int(attr.i)
        elif attr.type == 4:  # TENSOR
            value_dict[attr.name] = numpy_helper.to_array(attr.t)
            pass
        elif attr.type == 7:  # INTS
            value_dict[attr.name] = list(attr.ints)
        else:
            raise RuntimeError(f"unhandled attribute type {attr.type}")
    return value_dict



"""Module for formulating a :external+sklearn:py:class:`sklearn.svm.SVR`,
:external+sklearn:py:class:`sklearn.svm.SVC`,
:external+sklearn:py:class:`sklearn.svm.LinearSVR`, or
:external+sklearn:py:class:`sklearn.svm.LinearSVC` into a PySCIPOpt Model.
"""

import numpy as np
from sklearn.svm import LinearSVC, LinearSVR

from ..exceptions import NoModel
from ..modelling import AbstractPredictorConstr
from ..modelling.classification import argmax_bound_formulation
from .skgetter import SKgetter


def add_support_vector_regressor_constr(
    scip_model,
    support_vector_regressor,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate support_vector_regressor in scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    support_vector_regressor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    support_vector_regressor : :external+sklearn:py:class:`sklearn.svm.SVR`
        The support vector regressor to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for support vector in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for support vector in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    SupportVectorConstr
       Object containing information about what was added to scip_model to formulate
       support_vector_regressor.

    Note
    ----
    |VariablesDimensionsWarn|
    """

    return SupportVectorConstr(
        scip_model,
        support_vector_regressor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        False,
        **kwargs,
    )


def add_support_vector_classifier_constr(
    scip_model,
    support_vector_classifier,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate support_vector_classifier in scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    support_vector_classifier.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    support_vector_classifier : :external+sklearn:py:class:`sklearn.svm.SVC`
        The support vector classifier to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for support vector in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for support vector in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    SupportVectorConstr
       Object containing information about what was added to scip_model to formulate
       support_vector_classifier.

    Note
    ----
    |VariablesDimensionsWarn|
    """

    return SupportVectorConstr(
        scip_model,
        support_vector_classifier,
        input_vars,
        output_vars,
        unique_naming_prefix,
        True,
        **kwargs,
    )


class SupportVectorConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained :external+sklearn:py:class:`sklearn.svm.SVR`,
    :external+sklearn:py:class:`sklearn.svm.SVC`,
    :external+sklearn:py:class:`sklearn.svm.LinearSVR`, or
    :external+sklearn:py:class:`sklearn.svm.LinearSVC` with SCIP

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        classification,
        **kwargs,
    ):
        self.classification = classification
        if self.classification:
            if len(predictor.classes_) <= 2:
                self.output_size = 1
            else:
                raise NoModel(predictor, "Multi-classification is not supported for SVC")
        else:
            self.output_size = 1
        if isinstance(predictor, LinearSVR) or isinstance(predictor, LinearSVC):
            self.kernel = "linear"
        else:
            self.kernel = predictor.kernel
            if self.kernel not in ["linear", "poly"]:
                raise NoModel(predictor, f"Kernel type {self.kernel} not linear nor poly")
            if predictor.class_weight is not None:
                raise NoModel(predictor, "Non uniform class weights are not supported.")
        SKgetter.__init__(self, predictor, **kwargs)
        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Predict output variables y from input variables X using the support vectors.

        Both X and y should be arrays or lists of variables of conforming dimensions.
        """

        n_samples = self.input.shape[0]
        n_features = self.input.shape[-1]

        # Extract data from the predictor object
        intercept = self.predictor.intercept_[0]
        if self.kernel == "linear":
            # LinearSVM has a different dimension output to SVM
            if self.predictor.coef_.ndim > 1:
                coefs = self.predictor.coef_[0]
            else:
                coefs = self.predictor.coef_
        else:
            degree = self.predictor.degree
            n_support = np.sum(self.predictor.n_support_)
            dual_coefs = self.predictor.dual_coef_
            support_vectors = self.predictor.support_vectors_
            gamma = self.predictor._gamma  # TODO: Is this possible to access otherwise?

        # Create additional variables for classification
        if self.classification:
            regression_vars = self.scip_model.addMatrixVar(
                (n_samples, 1),
                vtype="C",
                lb=None,
                ub=None,
                name=self.unique_naming_prefix + "_reg",
            )

        # Create the regression constraints for the SVR / SVC
        regression_cons = np.zeros((n_samples,), dtype=object)
        for i in range(n_samples):
            if not self.classification:
                out_var = self.output[i][0]
            else:
                out_var = regression_vars[i][0]
            if self.kernel == "linear":
                name = self.unique_naming_prefix + f"reg_sv_lin_"
                regression_cons[i] = self.scip_model.addCons(
                    sum(self.input[i][j] * coefs[j] for j in range(n_features)) + intercept
                    == out_var,
                    name=name,
                )
            else:
                name = self.unique_naming_prefix + f"reg_sv_poly_{i}"
                dot_products = [
                    sum(self.input[i][j] * support_vectors[s][j] for j in range(n_features))
                    for s in range(n_support)
                ]
                regression_cons[i] = self.scip_model.addCons(
                    sum(
                        dual_coefs[0][s] * (gamma**degree) * dot_products[s] ** degree
                        for s in range(n_support)
                    )
                    + intercept
                    == out_var,
                    name=name,
                )

        # In the case of classification add conversion to binaries (centres at 0)
        if self.classification:
            argmax_vars, argmax_cons = argmax_bound_formulation(
                self.scip_model,
                regression_vars,
                self.output,
                self.unique_naming_prefix,
                one_dim_center=0,
            )
            self._created_vars.append(regression_vars)
            for new_vars in argmax_vars:
                self._created_vars.append(new_vars)
            for new_cons in argmax_cons:
                self._created_cons.append(new_cons)

        self._created_cons.append(regression_cons)



"""Utility function to find function that add a predictor in dictionary."""


def get_convertor(predictor, convertors):
    """Return the convertor for a given predictor."""
    convertor = None
    try:
        convertor = convertors[type(predictor)]
    except KeyError:
        pass
    if convertor is None:
        for parent in type(predictor).mro():
            try:
                convertor = convertors[parent]
                break
            except KeyError:
                pass
    if convertor is None:
        name = type(predictor).__name__
        try:
            convertor = convertors[name]
        except KeyError:
            pass
    return convertor



""" Utilities for modelling decision trees """

import numpy as np
from pyscipopt import quicksum


def compute_leafs_bounds(tree, epsilon, infinity):
    """Compute the bounds that define each leaf of the tree"""
    node_count = tree["node_count"]
    n_features = tree["n_features"]
    children_left = tree["children_left"]
    children_right = tree["children_right"]
    feature = tree["feature"]
    threshold = tree["threshold"]

    node_lb = -np.ones((n_features, node_count)) * infinity
    node_ub = np.ones((n_features, node_count)) * infinity

    stack = [
        0,
    ]

    while len(stack) > 0:
        node = stack.pop()
        left = children_left[node]
        if left < 0:
            continue
        right = children_right[node]
        assert left not in stack
        assert right not in stack
        node_ub[:, right] = node_ub[:, node]
        node_lb[:, right] = node_lb[:, node]
        node_ub[:, left] = node_ub[:, node]
        node_lb[:, left] = node_lb[:, node]

        node_ub[feature[node], left] = threshold[node] - (epsilon / 2)
        node_lb[feature[node], right] = threshold[node] + (epsilon / 2)
        stack.append(right)
        stack.append(left)

    return node_lb, node_ub


def leaf_formulation(
    scip_model, _input, output, tree, unique_naming_prefix, epsilon, classification=False
):
    """Formulate decision tree using 'leaf' formulation

    We have one variable per leaf of the tree and a series of indicator constraints to
    define when that leaf is reached.

    The first step of the procedure is to derive input bounds for each leaf of the decision tree. These bounds will
    dictate for which input values the leaf can be reached. For a single sample, let x \reals^{nI} be the input,
    z {0,1}^{nL} be binary variables representing if a leaf is reached or not, and y \reals^{nO} be the output.

    .. math::

        \begin{align*}
        z_{i} -> x_{j} \geq leaf_lb[i][j] \forall i \in nL, j \in nI
        z_{i} -> x_{j} \leq leaf_ub[i][j] \forall i \in nL, j \in nI
        if classification:
            y_{j} == \sum z_{i} \forall i \in nL \text{, where class j is output of z_i}
            \sum y_{k} == 1
        else:
            z_{i} -> y_{k} = leaf_value[i][j] \forall i \in nL, k \in nO
        \sum z_{i} == 1
        \end{align*}

    """

    # Create names for items we want to access frequently
    n_samples = _input.shape[0]
    n_features = tree["n_features"]
    outdim = output.shape[-1]

    # Collect leaf nodes
    leaf_ids = tree["children_left"] <= -1
    n_leafs = sum(leaf_ids)
    leaf_vars = scip_model.addMatrixVar(
        (n_samples, n_leafs), vtype="B", lb=0, ub=None, name=unique_naming_prefix + "leaf"
    )

    # Calculate bounds for each leaf node
    (node_lb, node_ub) = compute_leafs_bounds(tree, epsilon, scip_model.infinity())

    # Create empty constraint objects
    output_class_sum_leaf_cons = np.zeros((n_samples, outdim), dtype=object)
    indicator_output_cons = np.zeros((n_samples, n_leafs, outdim, 2), dtype=object)
    indicator_leaf_lb = np.zeros((n_samples, n_leafs, n_features), dtype=object)
    indicator_leaf_ub = np.zeros((n_samples, n_leafs, n_features), dtype=object)

    # Iterate over all leaf nodes (They are the non-zero entries in leaf_ids)
    for i in range(n_samples):
        leafs_per_class = [0 for _ in range(outdim)]
        for j, node in enumerate(leaf_ids.nonzero()[0]):
            fixed_var = False
            # Fix the leaf variable to 0 if the input bounds do not allow the leaf to be reached
            for feature in range(n_features):
                if (
                    _input[i][feature].getLbOriginal() > node_ub[feature][node]
                    or _input[i][feature].getUbOriginal() < node_lb[feature][node]
                ):
                    scip_model.fixVar(leaf_vars[i][j], 0)
                    fixed_var = True
                    break
            # If the leaf could be reached, then add two sets of indicator constraints.
            # The first will enforce that a leaf node is only selected if the input values result in such a leaf.
            # The second force the appropriate value output by the leaf to be selected
            if not fixed_var:
                for feature in range(n_features):
                    name_lb = unique_naming_prefix + f"indicator_lb_{i}_{j}_{feature}"
                    name_ub = unique_naming_prefix + f"indicator_ub_{i}_{j}_{feature}"
                    feat_lb = node_lb[feature, node]
                    feat_ub = node_ub[feature, node]
                    if (
                        feat_lb > -scip_model.infinity()
                        and _input[i][feature].getLbOriginal() < feat_lb
                    ):
                        indicator_leaf_lb[i][j][feature] = scip_model.addConsIndicator(
                            -_input[i][feature] <= -feat_lb, leaf_vars[i][j], name=name_lb
                        )
                    if (
                        feat_ub < scip_model.infinity()
                        and _input[i][feature].getUbOriginal() > feat_ub
                    ):
                        indicator_leaf_ub[i][j][feature] = scip_model.addConsIndicator(
                            _input[i][feature] <= feat_ub, leaf_vars[i][j], name=name_ub
                        )
                # Iterate over the final output shape (num_outputs)
                # In the case of classification (num_classes), simply force the most frequent class to be selected
                if classification:
                    value = int(np.argmax(tree["value"][node][0]))
                    if outdim == 1:
                        if value == 1:
                            leafs_per_class[0] += leaf_vars[i][j]
                    else:
                        leafs_per_class[value] += leaf_vars[i][j]
                else:
                    for k in range(outdim):
                        name_ub = unique_naming_prefix + f"indicator_output_{i}_{j}_{k}_0"
                        name_lb = unique_naming_prefix + f"indicator_output_{i}_{j}_{k}_1"
                        value = tree["value"][node][k][0]
                        indicator_output_cons[i][j][k][0] = scip_model.addConsIndicator(
                            output[i][k] <= value, leaf_vars[i][j], name=name_ub
                        )
                        indicator_output_cons[i][j][k][1] = scip_model.addConsIndicator(
                            -output[i][k] <= -value, leaf_vars[i][j], name=name_lb
                        )
        # Add constraints that ensure the correct class is selected depending on the leaf
        if classification:
            for j in range(outdim):
                name = f"class_leaf_{i}_{j}"
                output_class_sum_leaf_cons[i][j] = scip_model.addCons(
                    output[i][j] == leafs_per_class[j], name=name
                )

    # Now add the constraints that only one leaf can be selected.
    # In the case of classification there is an additional constraint that only one class can be selected
    leaf_sum_cons = np.zeros(n_samples, dtype=object)
    for i in range(n_samples):
        name = unique_naming_prefix + f"sum_leafs_{i}"
        leaf_sum_cons[i] = scip_model.addCons(
            quicksum(leaf_vars[i][j] for j in range(leaf_vars.shape[-1])) == 1, name=name
        )

    # Finally set potentially stronger global bounds on the output variables (in the case of regression)
    if not classification:
        max_vals = [np.max(tree["value"][:, j, :]) for j in range(outdim)]
        min_vals = [np.min(tree["value"][:, j, :]) for j in range(outdim)]
        for i in range(n_samples):
            for j in range(outdim):
                if output[i][j].getLbOriginal() < min_vals[j]:
                    scip_model.chgVarLb(output[i][j], min_vals[j])
                if output[i][j].getUbOriginal() > max_vals[j]:
                    scip_model.chgVarUb(output[i][j], max_vals[j])

    # Now return the added constraints and variables
    if classification:
        return [leaf_vars], [
            indicator_leaf_lb,
            indicator_leaf_ub,
            output_class_sum_leaf_cons,
            leaf_sum_cons,
        ]
    else:
        return [leaf_vars], [
            indicator_leaf_lb,
            indicator_leaf_ub,
            indicator_output_cons,
            leaf_sum_cons,
        ]



"""Bases classes for modeling neural network layers."""

from ..base_predictor_constraint import AbstractPredictorConstr
from .activations import (
    add_identity_activation_constraint_layer,
    add_relu_activation_constraint_layer,
    add_sigmoid_activation_constraint_layer,
    add_softmax_activation_constraint_layer,
    add_softplus_activation_constraint_layer,
    add_tanh_activation_constraint_layer,
)
from .layers import ActivationLayer, DenseLayer


class BaseNNConstr(AbstractPredictorConstr):
    """Base class for inserting a neural network into SCIP."""

    def __init__(
        self, scip_model, predictor, input_vars, output_vars, unique_naming_prefix, **kwargs
    ):
        self.predictor = predictor
        self.act_dict = {
            "relu": add_relu_activation_constraint_layer,
            "identity": add_identity_activation_constraint_layer,
            "tanh": add_tanh_activation_constraint_layer,
            "logistic": add_sigmoid_activation_constraint_layer,
            "softmax": add_softmax_activation_constraint_layer,
            "softplus": add_softplus_activation_constraint_layer,
        }
        self._layers = []

        super().__init__(scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs)

    def __iter__(self):
        return self._layers.__iter__()

    def add_dense_layer(
        self,
        input_vars,
        layer_coefs,
        layer_intercept,
        activation,
        output_vars=None,
        unique_naming_prefix="",
        **kwargs,
    ):
        """Add a dense layer to PySCIPOpt Model.

        Parameters
        ----------

        input_vars : np.ndarray
            Decision variables used as input for predictor in model.
        layer_coefs:
            Coefficient for each node in a layer
        layer_intercept:
            Intercept bias
        activation: str
            Type of activation used in layer
        output_vars : np.ndarray or None, optional
            Output variables
        unique_naming_prefix : str, optional
            A unique naming prefix that is used before all variable and constraint names.
        """
        layer = DenseLayer(
            self.scip_model,
            input_vars,
            layer_coefs,
            layer_intercept,
            output_vars,
            activation,
            unique_naming_prefix,
            **kwargs,
        )
        self._layers.append(layer)
        return layer

    def add_activation_layer(
        self, input_vars, activation, activation_vars=None, unique_naming_prefix="", **kwargs
    ):
        """Add an activation layer to the SCIP model.

        Parameters
        ----------

        input_vars : np.ndarray
            Decision variables used as input for predictor in SCIP model.
        activation: str
            Type of activation function used in layer
        activation_vars : np.ndarray, optional
            Output variables
        unique_naming_prefix : str, optional
            A unique naming prefix that is used before all variable and constraint names.
        """
        layer = ActivationLayer(
            self.scip_model,
            activation_vars,
            input_vars,
            activation,
            unique_naming_prefix,
            **kwargs,
        )
        self._layers.append(layer)
        return layer



"""Implements some utility tools for all lightgbm objects."""

import numpy as np
from sklearn.base import is_classifier

from ..exceptions import NoModel, NoSolution, ParameterError
from ..modelling import AbstractPredictorConstr


class LGBgetter(AbstractPredictorConstr):
    """Utility class for lightgbm models convertors.

    Implement some common functionalities: check predictor is fitted, output dimension, get error

    Attributes
    ----------
    predictor
        Lightgbm predictor embedded into SCIP model.
    """

    def __init__(self, predictor, input_vars, output_type="regular", **kwargs):
        if not hasattr(predictor, "booster_"):
            raise ParameterError(
                "LightGBM model has not yet been fitted. There is nothing to model."
            )
        if predictor.boosting_type not in ["gbdt", "rf"]:
            raise NoModel(
                predictor,
                f"There is only support for LightGBM boosting type gbdt and rf. "
                f"Not {predictor.boosting_type}",
            )
        self.predictor = predictor

    def extract_raw_data_and_create_tree_vars(self, epsilon=0.0):
        """
        Function for extracting information from lgb._Booster and creating additional modelling variables.

        Parameters
        ----------
        epsilon : float, optional
            Small value used to impose strict inequalities for splitting nodes in
            MIP formulations.

        Returns
        -------
        trees : list
            A list of tree dictionaries similar in structure to that provided by SKlearn

        tree_vars : np.ndarray
            A numpy array filled with variables that represent output of trees from the trained LightGBM model
        """

        n_samples = self.input.shape[0]
        outdim = self.output.shape[1]

        # Extract the raw data
        raw = self.predictor.booster_.dump_model()
        n_features = self.predictor.n_features_
        n_trees = len(raw["tree_info"])
        n_estimators = self.predictor.n_estimators_

        def read_node(tree, tree_structure):
            # Increase the node count and keep track of the current node index
            node_idx = tree["node_count"]
            tree["node_count"] += 1

            # Add features specific to the node of the tree
            if "split_feature" in tree_structure:
                tree["feature"].append(tree_structure["split_feature"])
            else:
                tree["feature"].append(-1)
            if "threshold" in tree_structure:
                tree["threshold"].append(tree_structure["threshold"])
            else:
                tree["threshold"].append(-1)
            if "internal_value" in tree_structure:
                tree["value"].append([[tree_structure["internal_value"]]])
            else:
                tree["value"].append([[tree_structure["leaf_value"]]])
            if "decision_type" in tree_structure and tree_structure["decision_type"] != "<=":
                raise ParameterError(
                    f"Currently no support for LightGBM trees with decision type "
                    f"{tree_structure['decision_type']}"
                )

            # Add in dummy left and right children
            tree["children_left"].append(-1)
            tree["children_right"].append(-1)
            # Now recursively call a new node
            if "left_child" in tree_structure:
                tree["children_left"][node_idx] = tree["node_count"]
                tree = read_node(tree, tree_structure["left_child"])
            if "right_child" in tree_structure:
                tree["children_right"][node_idx] = tree["node_count"]
                tree = read_node(tree, tree_structure["right_child"])

            return tree

        trees = [
            [
                {
                    "node_count": 0,
                    "n_features": n_features,
                    "children_left": [],
                    "children_right": [],
                    "feature": [],
                    "threshold": [],
                    "value": [],
                }
                for _ in range(outdim)
            ]
            for _ in range(n_estimators)
        ]
        trees_converted = [0 for _ in range(outdim)]
        for i in range(n_trees):
            tree_raw = raw["tree_info"][i]["tree_structure"]
            class_idx = i % outdim
            tree_idx = trees_converted[class_idx]
            trees[tree_idx][class_idx] = read_node(trees[tree_idx][class_idx], tree_raw)
            trees[tree_idx][class_idx]["children_left"] = np.array(
                trees[tree_idx][class_idx]["children_left"], dtype=np.int32
            )
            trees[tree_idx][class_idx]["children_right"] = np.array(
                trees[tree_idx][class_idx]["children_right"], dtype=np.int32
            )
            trees[tree_idx][class_idx]["feature"] = np.array(
                trees[tree_idx][class_idx]["feature"], dtype=np.int32
            )
            trees[tree_idx][class_idx]["threshold"] = np.array(
                trees[tree_idx][class_idx]["threshold"], dtype=np.float64
            )
            trees[tree_idx][class_idx]["value"] = np.array(
                trees[tree_idx][class_idx]["value"], dtype=np.float64
            )
            trees_converted[class_idx] += 1

        shape = (n_samples, n_estimators, outdim)
        tree_vars = self.scip_model.addMatrixVar(shape, vtype="C", lb=None, ub=None, name="tree")

        return trees, tree_vars

    def get_error(self, eps=None):
        """
        Returns error in SCIP's solution with respect to the actual output of the trained predictor

        Parameters
        ----------
        eps : float or int or None, optional
            The maximum allowed tolerance for a mismatch between the actual predictive model and SCIP.
            If the error is larger than eps an appropriate warning is printed

        Returns
        -------
        error: np.ndarray
            The absolute values of the difference between SCIP's solution and the trained ML model's output given
            the input as defined by SCIP. The matrix is the same dimension as the output of the trained predictor.
            Using sklearn / pyscipopt, the absolute difference between model.predict(input) and scip.getVal(output).

        Raises
        ------
        NoSolution
            If SCIP has no solution (either was not optimized or is infeasible).
        """

        if self._has_solution:
            if not is_classifier(self.predictor):
                lgb_output_values = self.predictor.predict(self.input_values).reshape(
                    self.input.shape[0], self.output.shape[-1]
                )
            else:
                lgb_class_prediction = self.predictor.predict(self.input_values)
                lgb_output_values = np.zeros((self.input.shape[0], self.output.shape[-1]))
                for i, class_pred in enumerate(lgb_class_prediction):
                    if self.output.shape[-1] == 1:
                        lgb_output_values[i][0] = class_pred
                    else:
                        lgb_output_values[i][class_pred] = 1
            scip_output_values = self.output_values
            error = np.abs(lgb_output_values - scip_output_values)
            max_error = np.max(error)
            if eps is not None and max_error > eps:
                print(
                    f"SCIP output values of ML model {self.predictor} have larger than max error {max_error} > {eps}"
                )
            return error

        raise NoSolution()



import os

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from pyscipopt import Model, quicksum
from torch.utils.data import DataLoader
from torchvision import datasets, transforms

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view someone trying to fool a trained predictor for a given example.
This is referred to in the literature as an adversarial example.

We have access to open source data (thanks to: @article{lecun1998gradient,
  title={Gradient-based learning applied to document recognition},
  author={LeCun, Yann and Bottou, L{\'e}on and Bengio, Yoshua and Haffner, Patrick},
  journal={Proceedings of the IEEE},
  volume={86},
  number={11},
  pages={2278--2324},
  year={1998},
  publisher={Ieee}
})

This data is the famous MNIST dataset, which contains images of individual numbers that have been
drawn by children. Each image contains exactly one centred image, and is available everywhere as the
standard deep learning data set.

The goal of this MIP is to take an image from the dataset, and perturb it such that
the predictor now misclassifies it. There is a budget on how much the image can be perturbed.

Let I be the x coordinate and J be the y coordinate of the image.
Let f be the ML predictor that outputs the probability of each label for the input image
Let image[i][j] be the input image in grey scale with pixel values between 0-1.
Let x[i][j] the perturbed image in grey scale with pixel values between 0-1.
Let y be the probability of each label given the ML predictor
W.l.o.g assume that y[0] is the true label and y[1] is the fake label.

x[i][j] - image[i][j] <= abs_diff[i][j] for all i, j
image[i][j] - x[i][j] <= abs_diff[i][j] for all i, j
\sum_i,j abs_diff[i][j] <= 5
y = f(x)

max(y[1] - y[0])
"""


def build_and_optimise_adversarial_mnist_torch(
    data_seed=42,
    training_seed=42,
    n_pixel_1d=16,
    layer_size=16,
    n_layers=2,
    test=True,
    formulation="sos",
    build_only=False,
):
    # Set random seed for reproducibility and select the image that is going to be perturbed
    data_random_state = np.random.RandomState(data_seed)
    image_number = data_random_state.randint(low=0, high=30000)
    torch.manual_seed(training_seed)

    # Define transformations for the MNIST dataset
    transform = transforms.Compose(
        [
            transforms.Resize(n_pixel_1d),  # Resize the image
            transforms.ToTensor(),  # Convert images to tensors. This automatically normalises to [0,1]
        ]
    )

    # Get MNIST digit recognition data set
    train_dataset = datasets.MNIST(
        root="./tests/data/MNIST", train=True, transform=transform, download=True
    )
    test_dataset = datasets.MNIST(
        root="./tests/data/MNIST", train=False, transform=transform, download=True
    )

    # Create DataLoader for handling batches
    batch_size = 64
    train_dataloader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
    test_dataloader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)

    # Create the neural network
    layers = [nn.Flatten(), nn.Linear(n_pixel_1d**2, layer_size), nn.ReLU()]
    nn.init.xavier_uniform_(layers[1].weight)
    for i in range(n_layers - 1):
        layers.append(nn.Linear(layer_size, layer_size))
        nn.init.xavier_uniform_(layers[-1].weight)
        layers.append(nn.ReLU())
    layers.append(nn.Linear(layer_size, 10))
    nn.init.xavier_uniform_(layers[-1].weight)
    reg = nn.Sequential(*layers)
    nn.utils.clip_grad_norm_(reg.parameters(), 0.1)

    # If the model is already saved then skip the training step
    saved_neural_network_path = "./tests/data/adversarial.pt"
    if test and os.path.isfile(saved_neural_network_path):
        reg.load_state_dict(torch.load(saved_neural_network_path))
        reg.eval()

    else:
        # Initialise the loss function, and optimizer
        criterion = nn.CrossEntropyLoss()
        optimizer = optim.Adam(reg.parameters(), weight_decay=0.001)

        # Training loop
        num_epochs = 20

        for epoch in range(num_epochs):
            reg.train()
            for batch_X, batch_y in train_dataloader:
                # Forward pass
                outputs = reg(batch_X)

                # Calculate loss
                loss = criterion(outputs, batch_y)

                # Backward pass and optimization
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            # Print training loss after each epoch
            print(f"Epoch [{epoch + 1}/{num_epochs}], Training Loss: {loss.item():.4f}")

        # Test the trained model on the test dataset
        reg.eval()
        correct = 0
        total = 0

        # Determine the accuracy of the trained model on the test dataset
        with torch.no_grad():
            for batch_X, batch_y in test_dataloader:
                outputs = reg(batch_X)
                _, predicted = torch.max(outputs, 1)
                total += batch_y.size(0)
                correct += (predicted == batch_y).sum().item()

        accuracy = correct / total
        print(f"Test Accuracy: {accuracy * 100:.2f}%")

        # Save the neural network for repeated tests
        if test:
            torch.save(reg.state_dict(), saved_neural_network_path)

    # Now create the MILP Model. The MILP will have a budget of pixels to change and decide which pixels and which new
    # values most change the classification of an image
    scip = Model()

    output_values = reg.forward(train_dataset[image_number][0])
    sorted_labels = torch.argsort(output_values)
    right_label = sorted_labels[0][-1]
    wrong_label = sorted_labels[0][-2]

    # Create the input variables
    input_vars = np.zeros((1, n_pixel_1d, n_pixel_1d), dtype=object)
    for i in range(n_pixel_1d):
        for j in range(n_pixel_1d):
            # Tight bounds are important for MIP formulations of neural networks. They often drastically improve
            # performance. As our data is a scaled image, it is in the range [0, 1].
            # These bounds will then propagate to other variables.
            input_vars[0][i][j] = scip.addVar(name=f"x_{i}_{j}", vtype="C", lb=0, ub=1)

    # Create the output variables. (Note that these variables will be automatically constructed if not specified)
    output_vars = np.zeros((10,), dtype=object)
    for i in range(10):
        output_vars[i] = scip.addVar(name=f"y_{i}", vtype="C", lb=None, ub=None)

    # Create the difference variables
    sum_max_diff = data_random_state.uniform(low=4.5, high=5.5)
    abs_diff = np.zeros((n_pixel_1d, n_pixel_1d), dtype=object)
    for i in range(n_pixel_1d):
        for j in range(n_pixel_1d):
            abs_diff[i][j] = scip.addVar(name=f"abs_diff_{i}", vtype="C", lb=0, ub=1)

    # Create constraints ensuring only a total certain amount of the picture can change
    for i in range(n_pixel_1d):
        for j in range(n_pixel_1d):
            scip.addCons(
                abs_diff[i][j] >= input_vars[0][i][j] - train_dataset[image_number][0][0][i][j]
            )
            scip.addCons(
                abs_diff[i][j] >= -input_vars[0][i][j] + train_dataset[image_number][0][0][i][j]
            )

    scip.addCons(
        quicksum(quicksum(abs_diff[i][j] for j in range(n_pixel_1d)) for i in range(n_pixel_1d))
        <= sum_max_diff
    )

    # Set an objective to maximise the difference between the correct and the wrong label
    scip.setObjective(-output_vars[wrong_label] + output_vars[right_label] + 1)

    # Add the ML constraint
    pred_cons = add_predictor_constr(
        scip,
        reg,
        input_vars,
        output_vars,
        unique_naming_prefix="adversarial_",
        formulation=formulation,
    )

    if not build_only:
        scip.optimize()

        # We can check the "error" of the MIP embedding by determining the difference between the Torch and SCIP output
        if np.max(pred_cons.get_error()) > 10**-3:
            error = np.max(pred_cons.get_error())
            raise AssertionError(f"Max error {error} exceeds threshold of {10**-3}")

    return scip


def test_mnist_torch():
    scip = build_and_optimise_adversarial_mnist_torch(
        data_seed=42,
        training_seed=42,
        n_pixel_1d=12,
        layer_size=10,
        n_layers=2,
        test=True,
        formulation="sos",
    )


def test_mnist_torch_bigm():
    scip = build_and_optimise_adversarial_mnist_torch(
        data_seed=42,
        training_seed=42,
        n_pixel_1d=12,
        layer_size=10,
        n_layers=2,
        test=True,
        formulation="bigm",
    )



import numpy as np
from pyscipopt import Model
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.neural_network import MLPRegressor
from utils import train_torch_neural_network

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view of a job scheduler. Specifically, we would like
to schedule jobs to individual compute cores, such that the minimum efficieny of any core
is maximised.

The model and data is thanks to:
@article{lombardi2017empirical,
  title={Empirical decision model learning},
  author={Lombardi, Michele and Milano, Michela and Bartolini, Andrea},
  journal={Artificial Intelligence},
  volume={244},
  pages={343--367},
  year={2017},
  publisher={Elsevier}
}
The data was accessed here: https://bitbucket.org/m_lombardi/eml-aij-2015-resources/src/master/training/

We have built a predictor that takes as input the average CPI (clocks per instruction or cycles per instruction)
of jobs assigned to a compute core, the average CPI of its neighbours, and the average CPI of cores far away.
The predictor outputs the efficiency of the core in how it handles the assigned jobs.

The goal of this MIP is to maximise the minimum efficiency over all cores. It achieves this by assigning
jobs to each compute core such that the predicted efficiency of the worst core is maximised.

Let I be the index set of jobs
Let K be the index set of compute cores
Let N[k] be some neighbourhood of other compute cores for compute core k
The variable x[i][k] defines if job i is assigned to compute core k. The variables are binary
cpi[i] is the clock per instructions of each job, and is input to the problem

The MIP model is as follows

min -y + 1

sum_k x[i][k] = 1 for all k
sum_i x[i][k] = |I| / |K| for all k
avg_cpi[k] = (|K| / |I|) sum_i cpi[i] * x[i][k] for all k
avg_neigh[k] = (1 / |N[k]|) * sum_k'_N[k] avg_cpi[k] for all k
avg_far[k] = (1 / (|K| - |N[k]|)) * sum_k'_N-N[k] avg_cpi[k] for all k
eff[k] = f(avg_cpi[k], avg_neigh[k], avg_far[k])
y <= eff[k] for all k
"""


def read_data_for_core(k):
    with open(f"tests/data/WDP/ts{k}.txt") as s:
        data_raw = s.readlines()

    x = np.zeros((len(data_raw) - 3, 3))
    y = np.zeros((len(data_raw) - 3,))

    for i, line in enumerate(data_raw[3:]):
        line = line.split("\n")[0].split(" ")
        assert len(line) == 5
        x[i][0] = float(line[0])
        x[i][1] = float(line[2])
        x[i][2] = float(line[3])
        y[i] = float(line[4])

    return x, y


def get_neighbours(k, n, m):
    neighbours = set()
    if k % m != 0:
        neighbours.add(k - 1)
    if (k + 1) % m != 0:
        neighbours.add(k + 1)
    if k - m >= 0:
        neighbours.add(k - m)
    if k + m <= n * m - 1:
        neighbours.add(k + m)

    not_neighbours = {i for i in range(n * m)}
    not_neighbours = not_neighbours - neighbours

    return sorted(list(neighbours)), sorted(list(not_neighbours))


def build_and_optimise_workload_dispatching(
    num_layers_or_estimators=4,
    layer_size_or_depth=3,
    nn_or_gbdt="gbdt",
    formulation="sos",
    framework="sklearn",
    training_seed=42,
    data_seed=42,
    build_only=False,
):
    n = 6
    m = 8
    n_jobs = 6 * n * m
    data_random_state = np.random.RandomState(data_seed)
    training_random_state = np.random.RandomState(training_seed)

    # Initialise the SCIP Model and all variables
    scip = Model()
    job_vars = np.zeros((n_jobs, n * m), dtype=object)
    avg_cpi_vars = np.zeros((n * m,), dtype=object)
    avg_neighbour_cpi_vars = np.zeros((n * m,), dtype=object)
    avg_far_cpi_vars = np.zeros((n * m,), dtype=object)
    efficiency_vars = np.zeros((n * m,), dtype=object)
    max_min_efficiency_var = scip.addVar(vtype="C", lb=0, ub=1, name="max_min_eff")

    for i in range(n * m):
        for j in range(n_jobs):
            job_vars[j][i] = scip.addVar(vtype="B", name=f"job_{j}_core_{i}")
        avg_cpi_vars[i] = scip.addVar(vtype="C", lb=-1, ub=0.5, name=f"avg_cpi_{i}")
        avg_neighbour_cpi_vars[i] = scip.addVar(vtype="C", lb=-1, ub=0.5, name=f"avg_n_cpi_{i}")
        avg_far_cpi_vars[i] = scip.addVar(vtype="C", lb=-1, ub=0.5, name=f"avg_f_cpi_{i}")
        efficiency_vars[i] = scip.addVar(vtype="C", lb=0, ub=1, name=f"eff_{i}")

    # Read the data and train + embed all ML predictors
    min_avg_cpi = np.inf
    max_avg_cpi = -np.inf
    predictor_constraints = []
    for i in range(n * m):
        x, y = read_data_for_core(i)
        min_avg_cpi = min(np.min(x[:, 0]), min_avg_cpi)
        max_avg_cpi = max(np.max(x[:, 0]), max_avg_cpi)
        layer_sizes = tuple([layer_size_or_depth for i in range(num_layers_or_estimators)])
        if nn_or_gbdt == "nn":
            if framework == "sklearn":
                reg = MLPRegressor(
                    hidden_layer_sizes=layer_sizes,
                    random_state=training_random_state,
                    learning_rate_init=0.01,
                ).fit(x, y)
            else:
                reg = train_torch_neural_network(
                    x, y, num_layers_or_estimators, layer_size_or_depth, training_seed
                )
        else:
            reg = GradientBoostingRegressor(
                n_estimators=num_layers_or_estimators,
                random_state=training_random_state,
                max_depth=layer_size_or_depth,
            ).fit(x, y)
        pred_input = [avg_cpi_vars[i], avg_neighbour_cpi_vars[i], avg_far_cpi_vars[i]]
        pred_cons = add_predictor_constr(
            scip,
            reg,
            pred_input,
            [efficiency_vars[i]],
            unique_naming_prefix=f"p_{i}_",
            formulation=formulation,
        )
        predictor_constraints.append(pred_cons)

    # Randomly generate an instance and create the appropriate constraints
    cpi = data_random_state.uniform(
        low=(min_avg_cpi - 0.01), high=(max_avg_cpi + 0.01), size=n_jobs
    )
    for i in range(n * m):
        # Add the average CPI constraint
        avg_cpi = sum(cpi[j] * job_vars[j][i] for j in range(n_jobs)) / 6
        scip.addCons(avg_cpi_vars[i] == avg_cpi, name=f"avg_cpi_cons_{i}")
        # Generate the neighbours of the core
        neighbours, not_neighbours = get_neighbours(i, n, m)
        # Add the average neighbour constraint
        neighbour_avg = sum(avg_cpi_vars[k] for k in neighbours) / (len(neighbours))
        scip.addCons(avg_neighbour_cpi_vars[i] == neighbour_avg, name=f"avg_cpi_neigh_cons_{i}")
        # Add the average far away core constraint
        not_neighbour_avg = sum(avg_cpi_vars[k] for k in not_neighbours) / len(not_neighbours)
        scip.addCons(avg_far_cpi_vars[i] == not_neighbour_avg, name=f"avg_far_cpi_cons_{i}")
        # Add the constraint that global efficiency is less than any single core efficiency
        scip.addCons(max_min_efficiency_var <= efficiency_vars[i], name=f"eff_bound_{i}")

    scip.setObjective(-max_min_efficiency_var + 1)

    if not build_only:
        scip.optimize()
        for pred_cons in predictor_constraints:
            if np.max(pred_cons.get_error()) > 10**-4:
                error = np.max(pred_cons.get_error())
                raise AssertionError(f"Max error {error} exceeds threshold of {10 ** -4}")

    return scip


def test_gbdt_workload_dispatching():
    build_and_optimise_workload_dispatching(
        nn_or_gbdt="gbdt", num_layers_or_estimators=4, layer_size_or_depth=2
    )


def test_nn_workload_dispatching():
    build_and_optimise_workload_dispatching(
        nn_or_gbdt="nn", num_layers_or_estimators=2, layer_size_or_depth=3
    )


def test_nn_workload_dispatching_bigm():
    build_and_optimise_workload_dispatching(
        nn_or_gbdt="nn", num_layers_or_estimators=2, layer_size_or_depth=3, formulation="bigm"
    )



import numpy as np
import pandas as pd
from pyscipopt import Model
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.svm import SVR
from utils import train_torch_neural_network

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view of the World Food Programme (WFP). We imagine a simplified
variant of their food delivery problem, where one must minimise costs while ensuring that an
appropriate amount of food is delivered. Appropriate is defined both by some
nutritional value threshold and by some palatable threshold.

The model is a subproblem of the larger optimisation problem by:
@article{peters2021nutritious,
  title={The nutritious supply chain: optimizing humanitarian food assistance},
  author={Peters, Koen and Silva, S{\'e}rgio
  and Gon{\c{c}}alves, Rui and Kavelj, Mirjana and Fleuren, Hein
  and den Hertog, Dick and Ergun, Ozlem and Freeman, Mallory},
  journal={INFORMS Journal on Optimization},
  volume={3},
  number={2},
  pages={200--226},
  year={2021},
  publisher={INFORMS}
}
The data is accessed thanks to:
@article{maragno2023mixed,
  title={Mixed-integer optimization with constraint learning},
  author={Maragno, Donato and Wiberg, Holly and
  Bertsimas, Dimitris and Birbil, {\c{S}} {\.I}lker and den Hertog, Dick and Fajemisin, Adejuyigbe O},
  journal={Operations Research},
  year={2023},
  publisher={INFORMS}
}
Accessed here: https://github.com/hwiberg/OptiCL/tree/main/notebooks/WFP/processed-data

We have built a predictor that takes as input a variety of features (in this case the
amount of each type of food), and determines how palatable of the collection of food.

The goal of this MIP is to minimise the transport cost of food to those in need while ensuring that
all nutritional requirements are met and that the general collection of food
is palatable.

Let I be the index set of the type of food
Let J be the index set of the type of nutrient
Let x[i] be the variables representing the amount of fod of type i that is purchased
Let c[i] be the transport cost of food type i
Let n[i][j] be the nutritional value of nutrient j for one unit of food type i
Let N[j] be the minimum nutritional requirement for nutrient j
Let t be a value in range [0,1] (0: horrible, 1: incredible)
Let f be the ML predictor for how palatable a set of food is

The MIP model is:

min sum_i c[i] * x[i]

sum_i n[i][j] * x[i] >= N[j] forall j
x_salt = 5
x_sugar = 20
f(x) >= t
"""


def build_and_optimise_palatable_diet(
    mlp_gbdt_svm="mlp",
    formulation="sos",
    framework="sklearn",
    n_estimators_or_layers=2,
    layer_sizes_or_depth=16,
    degree=2,
    training_seed=42,
    data_seed=42,
    build_only=False,
):
    # Initialise a random state
    data_random_state = np.random.RandomState(data_seed)
    training_random_state = np.random.RandomState(training_seed)

    # Read in all the data
    nutr_val = pd.read_excel(
        "tests/data/WFP/instance_data.xlsx",
        sheet_name="nutr_val",
        index_col="Food",
        engine="openpyxl",
    ).to_numpy()
    nutr_req = pd.read_excel(
        "tests/data/WFP/instance_data.xlsx",
        sheet_name="nutr_req",
        index_col="Type",
        engine="openpyxl",
    ).to_numpy()
    cost_p = (
        pd.read_excel(
            "tests/data/WFP/instance_data.xlsx",
            sheet_name="FoodCost",
            index_col="Supplier",
            engine="openpyxl",
        )
        .iloc[0, :]
        .to_numpy()
    )
    n_food = nutr_val.shape[0]
    n_nutrients = nutr_val.shape[1]
    data = pd.read_csv("tests/data/WFP/palatable_data.csv").sample(frac=1)
    y = data["label"].to_numpy()
    x = data.drop(["label"], axis=1, inplace=False).to_numpy()

    # Train the regression model on whether the diet is palatable or not
    if mlp_gbdt_svm == "mlp":
        if framework == "sklearn":
            hidden_layer_sizes = tuple(
                [layer_sizes_or_depth for i in range(n_estimators_or_layers)]
            )
            reg = MLPRegressor(
                random_state=training_random_state,
                hidden_layer_sizes=hidden_layer_sizes,
            ).fit(x, y.reshape(-1))
        else:
            reg = train_torch_neural_network(
                x, y, n_estimators_or_layers, layer_sizes_or_depth, training_seed, reshape=True
            )
    elif mlp_gbdt_svm == "gbdt":
        reg = GradientBoostingRegressor(
            max_depth=layer_sizes_or_depth,
            n_estimators=n_estimators_or_layers,
            random_state=training_random_state,
        ).fit(x, y)
    elif mlp_gbdt_svm == "svm":
        reg = SVR(kernel="poly", degree=degree).fit(x, y)
    else:
        raise ValueError(f"No known model for type {mlp_gbdt_svm}")

    # Create the SCIP Model and model all variables
    scip = Model()
    input_vars = np.zeros((1, n_food), dtype=object)
    for i in range(n_food):
        input_vars[0][i] = scip.addVar(vtype="C", lb=0, ub=10, name=f"x_{i}")
    output_vars = np.zeros((1, 1), dtype=object)
    output_vars[0][0] = scip.addVar(vtype="C", lb=0, ub=1, name="y")

    # Add the nutritional constraints
    for j in range(n_nutrients):
        scip.addCons(
            sum(nutr_val[i][j] * input_vars[0][i] for i in range(n_food)) >= nutr_req[0][j],
            name=f"nutr_req_{j}",
        )
    # Fix salt and sugar as they are always constant in the training data
    scip.addCons(input_vars[0][9] == 0.05, name="salt_const")
    scip.addCons(input_vars[0][20] == 0.2, name="sugar_const")

    # Set the objective to minimise costs
    scip.setObjective(sum(cost_p[i] * input_vars[0][i] for i in range(n_food)), sense="minimize")

    # Insert the ML predictor
    pred_cons = add_predictor_constr(
        scip,
        reg,
        input_vars,
        output_vars,
        unique_naming_prefix="reg_",
        epsilon=0.001,
        formulation=formulation,
    )

    # Add a minimum palatable constraint
    min_palatable = data_random_state.uniform(low=0.5, high=0.52)
    scip.addCons(output_vars[0][0] >= min_palatable, name="palatable")

    # Optimise the model!
    if not build_only:
        scip.optimize()
        if np.max(pred_cons.get_error()) > 10**-4:
            error = np.max(pred_cons.get_error())
            raise AssertionError(f"Max error {error} exceeds threshold of {10 ** -4}")

    return scip


def test_svm_palatable_diet():
    build_and_optimise_palatable_diet(mlp_gbdt_svm="svm", degree=1)


def test_nn_palatable_diet():
    build_and_optimise_palatable_diet(
        mlp_gbdt_svm="mlp", n_estimators_or_layers=2, layer_sizes_or_depth=16
    )


def test_nn_palatable_diet_bigm():
    build_and_optimise_palatable_diet(
        mlp_gbdt_svm="mlp", n_estimators_or_layers=2, layer_sizes_or_depth=16, formulation="bigm"
    )


def test_gbdt_palatable_diet():
    build_and_optimise_palatable_diet(
        mlp_gbdt_svm="gbdt", n_estimators_or_layers=5, layer_sizes_or_depth=4
    )



"""Implements some utility tools for all scikit-learn objects."""

import numpy as np
from sklearn.base import ClusterMixin, is_classifier
from sklearn.multioutput import _MultiOutputEstimator
from sklearn.utils.validation import check_is_fitted

from ..exceptions import NoSolution
from ..modelling import AbstractPredictorConstr


class SKgetter(AbstractPredictorConstr):
    """Utility class for sklearn models convertors.

    Implement some common functionalities: check predictor is fitted, get error

    Attributes
    ----------
    predictor
        Scikit-Learn predictor embedded into SCIP model.
    """

    def __init__(self, predictor, **kwargs):
        check_is_fitted(predictor)
        self.predictor = predictor

    def get_error(self, eps=None):
        """
        Returns error in SCIP's solution with respect to the actual output of the trained predictor

        Parameters
        ----------
        eps : float or int or None, optional
            The maximum allowed tolerance for a mismatch between the actual predictive model and SCIP.
            If the error is larger than eps an appropriate warning is printed

        Returns
        -------
        error: np.ndarray
            The absolute values of the difference between SCIP's solution and the trained ML model's output given
            the input as defined by SCIP. The matrix is the same dimension as the output of the trained predictor.
            Using sklearn / pyscipopt, the absolute difference between model.predict(input) and scip.getVal(output).

        Raises
        ------
        NoSolution
            If SCIP has no solution (either was not optimized or is infeasible).
        """

        if self._has_solution:
            if (
                not is_classifier(self.predictor) and not isinstance(self.predictor, ClusterMixin)
            ) or isinstance(self.predictor, _MultiOutputEstimator):
                sk_output_values = self.predictor.predict(self.input_values).reshape(
                    self.input.shape[0], self.output.shape[-1]
                )
            else:
                sk_class_prediction = self.predictor.predict(self.input_values)
                sk_output_values = np.zeros((self.input.shape[0], self.output.shape[-1]))
                for i, class_pred in enumerate(sk_class_prediction):
                    if self.output.shape[-1] == 1:
                        sk_output_values[i][0] = class_pred
                    else:
                        sk_output_values[i][class_pred] = 1
            scip_output_values = self.output_values
            error = np.abs(sk_output_values - scip_output_values)
            max_error = np.max(error)
            if eps is not None and max_error > eps:
                print(
                    f"SCIP output values of ML model {self.predictor} have larger than max error {max_error} > {eps}"
                )
            return error

        raise NoSolution()


class SKtransformer(AbstractPredictorConstr):
    """Utility class for sklearn preprocessing models convertors.

    Implement some common functionalities.

    Attributes
    ----------
    transformer
        Scikit-Learn transformer embedded into SCIP Model.
    """

    def __init__(
        self,
        scip_model,
        transformer,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        **kwargs,
    ):
        self.transformer = transformer
        check_is_fitted(transformer)
        super().__init__(scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs)
        # As transforming units can only ever be used as non-final stages of a pipeline, the output are created vars
        self._created_vars.append(self.output)

    def get_error(self, eps=None):
        """Returns error in SCIP's solution with respect to the actual output of the trained predictor

        Parameters
        ----------
        eps : float or int or None, optional
            The maximum allowed tolerance for a mismatch between the actual predictive model and SCIP.
            If the error is larger than eps an appropriate warning is printed

        Returns
        -------
        error: np.ndarray
            The absolute values of the difference between SCIP's solution and the trained ML model's output given
            the input as defined by SCIP. The matrix is the same dimension as the output of the fitted transformer.
            Using sklearn / pyscipopt, the absolute difference between
            transformer.transform(input) and scip.getVal(output).

        Raises
        ------
        NoSolution
            If SCIP has no solution (either was not optimized or is infeasible).
        """
        if self._has_solution:
            transformer = self.transformer
            input_values = self.input_values

            transformed = transformer.transform(input_values)
            if len(transformed.shape) == 1:
                transformed = transformed.reshape(-1, 1)

            r_val = np.abs(transformed - self.output_values)
            if eps is not None and np.max(r_val) > eps:
                print(f"{transformed} != {self.output_values}")
            return r_val

        raise NoSolution()



"""Internal module to make MIP modeling of activation functions."""

import numpy as np
from pyscipopt import exp, log, quicksum


def add_identity_activation_constraint_layer(layer, max_bound):
    """
    MIP model for identity activation on a layer

    Parameters
    ----------
    layer : AbstractNNLayer
        Layer to which activation is applied.

    max_bound : float or int
        The maximum bound for which propagation values will be stored

    Returns
    -------

    affine_cons : np.ndarray
        A numpy array containing the linear transformation constraints

    """

    n_samples = layer.input.shape[0]
    n_nodes_left = layer.input.shape[-1]
    n_nodes_right = layer.output.shape[-1]
    affine_cons = np.zeros((n_samples, n_nodes_right), dtype=object)

    # Perform some basic activity based bound propagation
    propagation_success, lbs, ubs = propagate_identity_bounds(
        layer, n_samples, n_nodes_left, n_nodes_right, False, max_bound
    )

    for i in range(n_samples):
        for j in range(n_nodes_right):
            rhs = (
                quicksum(layer.coefs[k][j] * layer.input[i][k] for k in range(n_nodes_left))
                + layer.intercept[j]
            )
            name = layer.unique_naming_prefix + f"affine_{i}_{j}"
            affine_cons[i][j] = layer.scip_model.addCons(layer.output[i][j] == rhs, name=name)
            # Propagate bounds
            if propagation_success:
                if abs(lbs[i][j]) < max_bound:
                    output_lb = layer.output[i][j].getLbOriginal()
                    layer.scip_model.chgVarLb(layer.output[i][j], max(lbs[i][j], output_lb))
                if abs(ubs[i][j]) < max_bound:
                    output_ub = layer.output[i][j].getUbOriginal()
                    layer.scip_model.chgVarUb(layer.output[i][j], min(ubs[i][j], output_ub))

    return affine_cons


def add_relu_activation_constraint_layer(layer, aux_vars, activation_only=True, formulation="sos"):
    """
    MIP model for ReLU activation on a layer

    Parameters
    ----------
    layer : AbstractNNLayer
        Layer to which activation is applied.

    aux_vars : np.ndarray
        Auxiliary variables that are used for the formulation. These are slack variables for the SOS formulation
        and binary activation variables for the Big-M formulation

    activation_only : bool, optional
        Whether this layer should only feature as an activation layer, i.e., skip the affine transformation

    formulation : str, optional
        The MIP formulation used for encoding the ReLU activation. Options are ["sos", "bigm"]

    Returns
    -------

    relu_cons : list
        A list of numpy arrays containing added constraints

    """

    # Initialise values for easy access and create empty constraint arrays
    n_samples = layer.input.shape[0]
    n_nodes_left = layer.input.shape[-1]
    n_nodes_right = layer.output.shape[-1]
    sos_cons = np.zeros((n_samples, n_nodes_right), dtype=object)
    cons_with_slack = np.zeros((n_samples, n_nodes_right), dtype=object)
    big_m_lb = np.zeros((n_samples, n_nodes_right), dtype=object)
    big_m_ub_inactive = np.zeros((n_samples, n_nodes_right), dtype=object)
    big_m_ub_active = np.zeros((n_samples, n_nodes_right), dtype=object)

    # Perform some basic activity based bound propagation
    max_bound = 10**5 if formulation == "sos" else np.inf
    propagation_success, lbs, ubs = propagate_identity_bounds(
        layer, n_samples, n_nodes_left, n_nodes_right, activation_only, layer.scip_model.infinity()
    )

    # Iterate over all nodes on the right hand side and create the appropriate constraints
    for i in range(n_samples):
        for j in range(n_nodes_right):
            # Propagate bounds
            if propagation_success:
                if abs(lbs[i][j]) < max_bound:
                    output_lb = layer.output[i][j].getLbOriginal()
                    layer.scip_model.chgVarLb(
                        layer.output[i][j], max(max(lbs[i][j], 0), output_lb)
                    )
                    if formulation == "sos":
                        layer.scip_model.chgVarUb(aux_vars[i][j], max(-lbs[i][j], 0))
                if abs(ubs[i][j]) < max_bound:
                    output_ub = layer.output[i][j].getUbOriginal()
                    layer.scip_model.chgVarUb(
                        layer.output[i][j], min(max(ubs[i][j], 0), output_ub)
                    )
                    if formulation == "sos":
                        layer.scip_model.chgVarLb(aux_vars[i][j], max(-ubs[i][j], 0))
            # Create layer constraints
            if layer.output[i][j].getLbOriginal() < 0:
                layer.scip_model.chgVarLb(layer.output[i][j], 0)
            if formulation == "sos":
                name = layer.unique_naming_prefix + f"slack_{i}_{j}"
                if activation_only:
                    cons_with_slack[i][j] = layer.scip_model.addCons(
                        layer.output[i][j] == layer.input[i][j] + aux_vars[i][j], name=name
                    )
                else:
                    rhs = quicksum(
                        layer.coefs[k][j] * layer.input[i][k] for k in range(n_nodes_left)
                    )
                    rhs += layer.intercept[j] + aux_vars[i][j]
                    cons_with_slack[i][j] = layer.scip_model.addCons(
                        layer.output[i][j] == rhs, name=name
                    )
                name = layer.unique_naming_prefix + f"sos_{i}_{j}"
                sos_cons[i][j] = layer.scip_model.addConsSOS1(
                    [layer.output[i][j], aux_vars[i][j]], name=name
                )
            else:
                if activation_only:
                    rhs = layer.input[i][j]
                else:
                    rhs = (
                        quicksum(
                            layer.coefs[k][j] * layer.input[i][k] for k in range(n_nodes_left)
                        )
                        + layer.intercept[j]
                    )
                name = layer.unique_naming_prefix + f"relu_lb_{i}_{j}"
                big_m_lb[i][j] = layer.scip_model.addCons(layer.output[i][j] >= rhs, name=name)
                name = layer.unique_naming_prefix + f"relu_ub_inactive_{i}_{j}"
                big_m_ub_inactive[i][j] = layer.scip_model.addCons(
                    layer.output[i][j] <= rhs - (1 - aux_vars[i][j]) * lbs[i][j], name=name
                )
                name = layer.unique_naming_prefix + f"relu_ub_active_{i}_{j}"
                big_m_ub_active[i][j] = layer.scip_model.addCons(
                    layer.output[i][j] <= aux_vars[i][j] * ubs[i][j], name=name
                )

    if formulation == "bigm":
        return [big_m_lb, big_m_ub_inactive, big_m_ub_active]
    else:
        return [cons_with_slack, sos_cons]


def add_sigmoid_activation_constraint_layer(layer, activation_only=True):
    """
    MIP model for Sigmoid activation on a layer

    Parameters
    ----------
    layer : AbstractNNLayer
        Layer to which activation is applied.

    activation_only : bool, optional
        Whether this layer should only feature as an activation layer, i.e., skip the affine transformation

    Returns
    -------

    sigmoid_cons : np.ndarray
        A numpy array containing added constraints

    """

    # Initialise values for easy access and create empty constraint arrays
    n_samples = layer.input.shape[0]
    n_nodes_left = layer.input.shape[-1]
    n_nodes_right = layer.output.shape[-1]
    sigmoid_cons = np.zeros((n_samples, n_nodes_right), dtype=object)

    # Iterate over all nodes on the right hand side and create the appropriate constraints
    for i in range(n_samples):
        for j in range(n_nodes_right):
            if layer.output[i][j].getLbOriginal() < 0:
                layer.scip_model.chgVarLb(layer.output[i][j], 0)
            if layer.output[i][j].getUbOriginal() > 1:
                layer.scip_model.chgVarUb(layer.output[i][j], 1)
            if activation_only:
                x = layer.input[i][j]
            else:
                x = (
                    quicksum(layer.coefs[k][j] * layer.input[i][k] for k in range(n_nodes_left))
                    + layer.intercept[j]
                )
            name = layer.unique_naming_prefix + f"sigmoid_{i}_{j}"
            sigmoid_cons[i][j] = layer.scip_model.addCons(
                layer.output[i][j] == 1 / (1 + exp(-x)), name=name
            )

    return sigmoid_cons


def add_tanh_activation_constraint_layer(layer, activation_only=True):
    """
    MIP model for tanh activation on a layer

    Parameters
    ----------
    layer : AbstractNNLayer
        Layer to which activation is applied.

    activation_only : bool, optional
        Whether this layer should only feature as an activation layer, i.e., skip the affine transformation

    Returns
    -------

    tanh_cons : np.ndarray
        A numpy array containing added constraints

    """

    # Initialise values for easy access and create empty constraint arrays
    n_samples = layer.input.shape[0]
    n_nodes_left = layer.input.shape[-1]
    n_nodes_right = layer.output.shape[-1]
    tanh_cons = np.zeros((n_samples, n_nodes_right), dtype=object)

    # Iterate over all nodes on the right hand side and create the appropriate constraints
    for i in range(n_samples):
        for j in range(n_nodes_right):
            if layer.output[i][j].getLbOriginal() < -1:
                layer.scip_model.chgVarLb(layer.output[i][j], -1)
            if layer.output[i][j].getUbOriginal() > 1:
                layer.scip_model.chgVarUb(layer.output[i][j], 1)
            if activation_only:
                x = layer.input[i][j]
            else:
                x = (
                    quicksum(layer.coefs[k][j] * layer.input[i][k] for k in range(n_nodes_left))
                    + layer.intercept[j]
                )
            name = layer.unique_naming_prefix + f"tanh_{i}_{j}"
            tanh_cons[i][j] = layer.scip_model.addCons(
                layer.output[i][j] == (1 - exp(-2 * x)) / (1 + exp(-2 * x)), name=name
            )

    return tanh_cons


def add_softmax_activation_constraint_layer(layer, activation_only=True):
    """
    MIP model for softmax activation on a layer

    Parameters
    ----------
    layer : AbstractNNLayer
        Layer to which activation is applied.

    activation_only : bool, optional
        Whether this layer should only feature as an activation layer, i.e., skip the affine transformation

    Returns
    -------

    softmax_cons : np.ndarray
        A numpy array containing added constraints

    """

    # Initialise values for easy access and create empty constraint arrays
    n_samples = layer.input.shape[0]
    n_nodes_left = layer.input.shape[-1]
    n_nodes_right = layer.output.shape[-1]
    softmax_cons = np.zeros((n_samples, n_nodes_right), dtype=object)

    # Iterate over all nodes on the right hand side and create the appropriate constraints
    for i in range(n_samples):
        sum_exp_expr = 0
        for j in range(n_nodes_right):
            if activation_only:
                sum_exp_expr += exp(layer.input[i][j])
            else:
                sum_exp_expr += exp(
                    quicksum(layer.coefs[k][j] * layer.input[i][k] for k in range(n_nodes_left))
                    + layer.intercept[j]
                )
        for j in range(n_nodes_right):
            if layer.output[i][j].getLbOriginal() < 0:
                layer.scip_model.chgVarLb(layer.output[i][j], 0)
            if layer.output[i][j].getUbOriginal() > 1:
                layer.scip_model.chgVarUb(layer.output[i][j], 1)
            if activation_only:
                x = layer.input[i][j]
            else:
                x = (
                    quicksum(layer.coefs[k][j] * layer.input[i][k] for k in range(n_nodes_left))
                    + layer.intercept[j]
                )
            name = layer.unique_naming_prefix + f"softmax_{i}_{j}"
            softmax_cons[i][j] = layer.scip_model.addCons(
                layer.output[i][j] == exp(x) / sum_exp_expr, name=name
            )

    return softmax_cons


def add_softplus_activation_constraint_layer(layer, activation_only=True):
    """
    MIP model for softplus activation on a layer

    Parameters
    ----------
    layer : AbstractNNLayer
        Layer to which activation is applied.

    activation_only : bool, optional
        Whether this layer should only feature as an activation layer, i.e., skip the affine transformation

    Returns
    -------

    softplus_cons : np.ndarray
        A numpy array containing added constraints

    """

    # Initialise values for easy access and create empty constraint arrays
    n_samples = layer.input.shape[0]
    n_nodes_left = layer.input.shape[-1]
    n_nodes_right = layer.output.shape[-1]
    softplus_cons = np.zeros((n_samples, n_nodes_right), dtype=object)

    # Iterate over all nodes on the right hand side and create the appropriate constraints
    for i in range(n_samples):
        for j in range(n_nodes_right):
            if layer.output[i][j].getLbOriginal() < 0:
                layer.scip_model.chgVarLb(layer.output[i][j], 0)
            if activation_only:
                x = layer.input[i][j]
            else:
                x = (
                    quicksum(layer.coefs[k][j] * layer.input[i][k] for k in range(n_nodes_left))
                    + layer.intercept[j]
                )
            name = layer.unique_naming_prefix + f"softplus_{i}_{j}"
            softplus_cons[i][j] = layer.scip_model.addCons(
                layer.output[i][j] == log(1 + exp(x)), name=name
            )

    return softplus_cons


def propagate_identity_bounds(
    layer, n_samples, n_nodes_left, n_nodes_right, activation_only, max_bound
):
    """
    Activity based bound propagation. Assume the worst case bound for each node individually and generate
    bounds for the next layer.

    Parameters
    ----------
    layer : AbstractNNLayer
        Layer to which activation is applied.
    n_samples : int
        The number of samples passed by the user
    n_nodes_left : int
        The number of nodes on the left (input) layer
    n_nodes_right : int
        The number of nodes on the right (output) layer
    activation_only : bool
        Whether the bounds are going to be used for an activation only layer. In this case no weighted sum is needed.
    max_bound : float or int
        The maximum absolute bound value for a variable before bound propagation termination

    Returns
    -------
    propagation : bool
        Whether propagation worked
    lbs : np.ndarray
        An array of worst-case scenario lower bounds
    ubs : np.ndarray
        An array of worst-case scenario upper bounds
    """

    ubs = np.zeros((n_samples, n_nodes_right))
    lbs = np.zeros(
        (
            n_samples,
            n_nodes_right,
        )
    )
    input_lbs = np.zeros((n_samples, n_nodes_left))
    input_ubs = np.zeros((n_samples, n_nodes_left))
    for i in range(n_samples):
        for k in range(n_nodes_left):
            input_lbs[i][k] = layer.input[i][k].getLbOriginal()
            input_ubs[i][k] = layer.input[i][k].getUbOriginal()

    # In the case of activation only we can simply return the input bounds
    if activation_only:
        return True, input_lbs, input_ubs

    # Skip propagation for the weighted sum if some bounds are larger than max_bound
    if np.max(np.abs(input_ubs)) + np.max(np.abs(input_lbs)) > max_bound:
        return False, ubs, lbs

    for i in range(n_samples):
        for j in range(n_nodes_right):
            ub = 0
            lb = 0
            for k in range(n_nodes_left):
                coefficient = layer.coefs[k][j]
                if coefficient > 0:
                    ub += input_ubs[i][k] * coefficient + 10**-6
                    lb += input_lbs[i][k] * coefficient - 10**-6
                elif coefficient < 0:
                    ub += input_lbs[i][k] * coefficient + 10**-6
                    lb += input_ubs[i][k] * coefficient - 10**-6
            ubs[i][j] = ub + layer.intercept[j] + 10**-6
            lbs[i][j] = lb + layer.intercept[j] - 10**-6

    return True, lbs, ubs



"""Module for formulating a :external+sklearn:py:class:`sklearn.ensemble.RandomForestRegressor`
or :external+sklearn:py:class:`sklearn.ensemble.RandomForestClassifier` into a PySCIPOpt Model.
"""
import numpy as np

from ..modelling import AbstractPredictorConstr
from ..modelling.gradient_boosting import (
    aggregate_estimator_outputs,
    create_aggregation_constraints,
    create_sklearn_tree_estimators,
)
from .skgetter import SKgetter


def add_random_forest_regressor_constr(
    scip_model,
    random_forest_regressor,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate random_forest_regressor in scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    random_forest_regressor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    random_forest_regressor : :external+sklearn:py:class:`sklearn.ensemble.RandomForestRegressor`
        The random forest regressor to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for random forest in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for random forest in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    RandomForestConstr
       Object containing information about what was added to scip_model to formulate
       random_forest_regressor.

    Note
    ----
    |VariablesDimensionsWarn|
    """

    return RandomForestConstr(
        scip_model,
        random_forest_regressor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        False,
        **kwargs,
    )


def add_random_forest_classifier_constr(
    scip_model,
    random_forest_classifier,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate random_forest_classifier in scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    random_forest_classifier.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    random_forest_classifier : :external+sklearn:py:class:`sklearn.ensemble.RandomForestClassifier`
        The random forest classifier to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for random forest in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for random forest in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    RandomForestConstr
       Object containing information about what was added to scip_model to formulate
       random_forest_classifier.

    Note
    ----
    |VariablesDimensionsWarn|
    """

    return RandomForestConstr(
        scip_model,
        random_forest_classifier,
        input_vars,
        output_vars,
        unique_naming_prefix,
        True,
        **kwargs,
    )


class RandomForestConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained :external+sklearn:py:class:`sklearn.ensemble.RandomForestRegressor`
    or :external+sklearn:py:class:`sklearn.ensemble.RandomForestClassifier` with SCIP

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        classification,
        **kwargs,
    ):
        self.estimators_ = []
        self.classification = classification
        if self.classification:
            if predictor.n_classes_ <= 2:
                self.output_size = 1
            else:
                self.output_size = predictor.n_classes_
        else:
            self.output_size = predictor.n_outputs_
        SKgetter.__init__(self, predictor, **kwargs)
        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Predict output variables y from input variables X using the decision tree.

        Both X and y should be arrays or lists of variables of conforming dimensions.
        """

        n_samples = self.input.shape[0]
        outdim = self.output.shape[-1]

        # Create individual models for each estimator (decision tree)
        estimators, tree_vars = create_sklearn_tree_estimators(
            self.scip_model,
            self.predictor,
            self.input,
            n_samples,
            outdim,
            self.unique_naming_prefix,
            self.classification,
            gbdt_or_rf="rf",
            **kwargs,
        )
        for estimator in estimators:
            self.estimators_.append(estimator)
        self._created_vars.append(tree_vars)

        # Get the constant shift used
        constant = np.array([0 for _ in range(self.output.shape[-1])])

        # Add the appropriate constraints for the case of regression and classification
        aggregate_tree_output = aggregate_estimator_outputs(tree_vars, 1, constant, aggr="avg")
        created_vars, created_cons = create_aggregation_constraints(
            self.scip_model,
            aggregate_tree_output,
            self.output,
            n_samples,
            outdim,
            self.unique_naming_prefix,
            self.classification,
        )
        for created_var in created_vars:
            self._created_vars.append(created_var)
        for created_con in created_cons:
            self._created_cons.append(created_con)



"""Module for formulating a :external+sklearn:py:class:`sklearn.pipeline.Pipeline`
into a PySCIPOpt Model.
The pipeline's transformers (or preprocessing steps) can be any of the following:
- :external+sklearn:py:class:`sklearn.preprocessing.StandardScaler`
- :external+sklearn:py:class:`sklearn.preprocessing.PolynomialFeatures`
- :external+sklearn:py:class:`sklearn.preprocessing.Normalizer`
- :external+sklearn:py:class:`sklearn.preprocessing.Binarizer`

The final step of the pipeline must be a valid predictor.

"""

from ..exceptions import NoModel
from ..modelling.base_predictor_constraint import AbstractPredictorConstr
from ..modelling.get_convertor import get_convertor
from ..registered_predictors import sklearn_convertors
from .preprocessing import sklearn_transformers
from .skgetter import SKgetter


def add_pipeline_constr(
    scip_model, pipeline, input_vars, output_vars=None, unique_naming_prefix="", **kwargs
):
    """Formulate pipeline into scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    pipeline.

    Parameters
    ----------
    scip_model : SCIP Model
        The PySCIPOpt Model where the predictor should be inserted.
    pipeline : :external+sklearn:py:class:`sklearn.pipeline.Pipeline`
        The pipeline to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for support vector in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for support vector in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    PipelineConstr
        Object containing information about what was added to scip_model to embed the
        predictor into it

    Raises
    ------
    NoModel
        If the translation to SCIP of one of the elements in the pipeline
        is not implemented or recognized.

    Notes
    -----
    |VariablesDimensionsWarn|
    """
    return PipelineConstr(
        scip_model, pipeline, input_vars, output_vars, unique_naming_prefix, **kwargs
    )


class PipelineConstr(SKgetter, AbstractPredictorConstr):
    """Class to formulate a trained :external+sklearn:py:class:`sklearn.pipeline.Pipeline`
    into a PySCIPOpt model.

    |ClassShort|
    """

    def __init__(
        self, scip_model, pipeline, input_vars, output_vars, unique_naming_prefix, **kwargs
    ):
        self._steps = []
        SKgetter.__init__(self, pipeline, **kwargs)
        AbstractPredictorConstr.__init__(
            self,
            scip_model,
            input_vars,
            output_vars,
            unique_naming_prefix,
            skip_validate=True,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        pipeline = self.predictor
        steps = self._steps
        input_vars = self._input

        for i, transformer in enumerate(pipeline[:-1]):
            convertor = get_convertor(transformer, sklearn_transformers())
            if convertor is None:
                raise NoModel(
                    self.predictor,
                    f"The sklearn pipeline transformer {transformer} is not supported",
                )
            steps.append(
                convertor(
                    self.scip_model,
                    transformer,
                    input_vars,
                    output_vars=None,
                    unique_naming_prefix=self.unique_naming_prefix + f"_step_{i}",
                    **kwargs,
                )
            )
            input_vars = steps[-1].output

        predictor = pipeline[-1]
        convertor = get_convertor(predictor, sklearn_convertors())
        if convertor is None:
            raise NoModel(
                self.predictor,
                f" The sklearn predictor {predictor} is not supported.",
            )
        steps.append(
            convertor(
                self.scip_model,
                predictor,
                input_vars,
                self._output,
                unique_naming_prefix=self.unique_naming_prefix + f"_pred_{len(steps) + 1}",
                **kwargs,
            )
        )
        if self._output is None:
            self._output = steps[-1].output
        self.output_size = steps[-1].output_size

        # As the output of the pipeline could not be validated before the individual steps were
        # created, we have to validate the model after creation.
        self._validate(self.input, self.output)

    @property
    def _has_solution(self):
        return self[-1]._has_solution

    @property
    def output(self):
        """Returns output variables of pipeline, i.e. output of its last step."""
        return self[-1].output

    @property
    def output_values(self):
        """Returns output values of pipeline in solution, i.e. output of its last step."""
        return self[-1].output_values

    @property
    def input(self):
        """Returns input variables of pipeline, i.e. input of its first step."""
        return self[0].input

    @property
    def input_values(self):
        """Returns input values of pipeline in solution, i.e. input of its first step."""
        return self[0].input_values

    def __getitem__(self, key):
        """Get an item from the pipeline steps."""
        return self._steps[key]

    def __iter__(self):
        """Iterate through pipeline steps."""
        return self._steps.__iter__()

    def __len__(self):
        """Get number of pipeline steps."""
        return self._steps.__len__()



"""Exceptions for pyscipopt-ml."""


class NotRegistered(Exception):
    """Predictor is not supported by pyscipopt-ml."""

    def __init__(self, predictor):
        super().__init__(
            f"Object of type {predictor} is not registered/supported with pyscipopt-ml"
        )


class NoModel(Exception):
    """No model is known for some structure."""

    def __init__(self, predictor, reason):
        if not isinstance(predictor, str):
            predictor = type(predictor).__name__
        super().__init__(f"Can't do model for {predictor}: {reason}")


class NoSolution(Exception):
    """SCIP doesn't have a solution."""

    def __init__(self):
        super().__init__("No solution available")


class ParameterError(Exception):
    """Wrong parameter to a function."""

    def __init__(self, message):
        super().__init__(message)



from .aggregate_tree_model import (
    aggregate_estimator_outputs,
    aggregated_estimator_formulation,
    create_aggregation_constraints,
    create_sklearn_tree_estimators,
    create_tree_estimators,
)



import numpy as np
from lightgbm import LGBMRegressor
from pyscipopt import Model, quicksum
from sklearn.ensemble import GradientBoostingRegressor, RandomForestRegressor
from sklearn.neural_network import MLPRegressor
from utils import read_csv_to_dict, train_torch_neural_network
from xgboost import XGBRegressor, XGBRFRegressor

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view of a wine manufacturer.

We have access to open source data (thanks to:
@article{cortez2009modeling,
  title={Modeling wine preferences by data mining from physicochemical properties},
  author={Cortez, Paulo and Cerdeira, Ant{\'o}nio and Almeida, Fernando and Matos, Telmo and Reis, Jos{\'e}},
  journal={Decision support systems},
  volume={47},
  number={4},
  pages={547--553},
  year={2009},
  publisher={Elsevier}
}
Data accessed here: https://www.kaggle.com/datasets/piyushgoyal443/red-wine-dataset)
for which we have built a predictor. The predictor takes as input a variety of wine features,
and determines the quality of the wine. This can either be modelled as a regression model,
with the quality being in the range [0,1], or some quality threshold can be set,
and the task made into a classification task.

The goal of this MIP is to create a diverse bouquet of wines with the highest average
quality. To do this, we make an assumption that the features of the grapes and resulting wine
are identical. In this MIP we can purchase grapes from a set of suppliers, all of whom
have grapes that by themselves would result in low quality wine. We can blend the grapes, however,
such that the predicted quality of the wine increases. Each supplier has a limit on the amount of
grapes that can be sold, and an associated cost per unit. Additionally,
we create the same amount of wine for each wine of the bouquet.

Let I be the index set of the wine bouquet that will be created
Let J be the index set of features for each grape / wine.
Let K be the index set of vineyards where the grapes can be purchased
Let x[i][j] be the feature j of wine i
Let y[i] be the quality of the wine i
Let m[i][k] be the amount of grapes from vineyard k used in wine i
let f be the ML predictor that determines the quality of a wine
Let feature_val[j][k] be the value of feature j from vineyard k
Let costs[k] be the cost of one unit of grapes from vineyard k
Let available[k] be the amount of grapes avalable from vineyard k
Let budget be the total budget available

The MIP model is:

x[i][j] = sum_k m[i][k] * feature_val[j][k] for all i, i
sum_k m[i][k] = 1 for all i
sum_i m[i][k] <= available[k]
y[i] = f(x[i][:]) for all i
sum_{i,k} m[i][k] * costs[k] <= budget

max(sum_i y[i])
"""


def build_and_optimise_wine_manufacturer(
    data_seed=42,
    training_seed=42,
    n_vineyards=35,
    n_wines_to_produce=5,
    framework="sklearn",
    gbdt_rf_or_mlp="rf",
    formulation="sos",
    n_estimators_layers=3,
    max_depth=3,
    layer_size=16,
    epsilon=0.0001,
    build_only=False,
):
    assert framework in ("sklearn", "xgboost", "lightgbm", "torch")
    assert gbdt_rf_or_mlp in ("gbdt", "rf", "mlp")
    data_random_state = np.random.RandomState(data_seed)
    training_random_state = np.random.RandomState(training_seed)

    # Path to red wine data
    data_dict = read_csv_to_dict("./tests/data/red_wine_quality.csv")

    features = [
        "fixed.acidity",
        "volatile.acidity",
        "citric.acid",
        "residual.sugar",
        "chlorides",
        "free.sulfur.dioxide",
        "total.sulfur.dioxide",
        "density",
        "pH",
        "sulphates",
        "alcohol",
    ]
    n_features = len(features)
    budget = 1.7 * n_wines_to_produce

    # Generate the actual input data arrays for the ML predictors
    X = []
    quality = np.array([float(x) for x in data_dict["quality"]]).reshape(
        -1,
    )
    for feature in features:
        X.append(np.array([float(x) for x in data_dict[feature]]))
    X = np.swapaxes(np.array(X), 0, 1)

    # Train the ML predictor
    if gbdt_rf_or_mlp == "mlp":
        if framework == "sklearn":
            hidden_layers = tuple([layer_size for i in range(n_estimators_layers)])
            reg = MLPRegressor(
                random_state=training_random_state, hidden_layer_sizes=hidden_layers
            ).fit(X, quality)
        else:
            reg = train_torch_neural_network(
                X, quality, n_estimators_layers, layer_size, training_seed, reshape=True
            )
    elif framework == "sklearn":
        if gbdt_rf_or_mlp == "rf":
            reg = RandomForestRegressor(
                random_state=training_random_state,
                n_estimators=n_estimators_layers,
                max_depth=max_depth,
            ).fit(X, quality)
        else:
            reg = GradientBoostingRegressor(
                random_state=training_random_state,
                n_estimators=n_estimators_layers,
                max_depth=max_depth,
            ).fit(X, quality)
    elif framework == "xgboost":
        if gbdt_rf_or_mlp == "gbdt":
            reg = XGBRegressor(
                random_state=training_random_state,
                n_estimators=n_estimators_layers,
                max_depth=max_depth,
            ).fit(X, quality)
        else:
            reg = XGBRFRegressor(
                random_state=training_random_state,
                n_estimators=n_estimators_layers,
                max_depth=max_depth,
            ).fit(X, quality)
    else:
        if gbdt_rf_or_mlp == "gbdt":
            reg = LGBMRegressor(
                random_state=training_random_state,
                n_estimators=n_estimators_layers,
                max_depth=max_depth,
            ).fit(X, quality)
        else:
            reg = LGBMRegressor(
                random_state=training_random_state,
                n_estimators=n_estimators_layers,
                max_depth=max_depth,
                boosting_type="rf",
                bagging_freq=1,
                bagging_fraction=0.5,
            ).fit(X, quality)

    # Create artificial data from some vineyards
    vineyard_order = np.arange(X.shape[0])
    data_random_state.shuffle(vineyard_order)
    vineyard_litre_limits = data_random_state.uniform(0.25, 0.35, n_vineyards)
    vineyard_costs = data_random_state.uniform(1, 2, n_vineyards)
    vineyard_features = []
    low_quality_vineyards_i = 0
    for i in vineyard_order:
        if low_quality_vineyards_i >= n_vineyards:
            break
        if quality[i] <= 5:
            low_quality_vineyards_i += 1
            vineyard_features.append(X[i])
    vineyard_features = np.array(vineyard_features)

    # Create the SCIP Model
    scip = Model()

    # Create variables deciding the features of each wine
    feature_vars = np.zeros((n_wines_to_produce, n_features), dtype=object)
    quality_vars = np.zeros((n_wines_to_produce, 1), dtype=object)
    wine_mixture_vars = np.zeros((n_wines_to_produce, n_vineyards), dtype=object)

    for i in range(n_wines_to_produce):
        quality_vars[i][0] = scip.addVar(vtype="C", lb=0, ub=10, name=f"quality_{i}")
        for j in range(n_features):
            max_val = np.max(X[:, j])
            min_val = np.min(X[:, j])
            lb = max(0, min_val - 0.1 * max_val)
            ub = 1.1 * max_val
            feature_vars[i][j] = scip.addVar(vtype="C", lb=lb, ub=ub, name=f"feature_{i}_{j}")
        for k in range(n_vineyards):
            wine_mixture_vars[i][k] = scip.addVar(
                vtype="C", lb=0, ub=vineyard_litre_limits[k], name=f"mixture_{i}_{k}"
            )

    # Now create constraints on the wine blending
    for i in range(n_wines_to_produce):
        for j in range(n_features):
            scip.addCons(
                feature_vars[i][j]
                == quicksum(
                    wine_mixture_vars[i][k] * vineyard_features[k][j] for k in range(n_vineyards)
                ),
                name=f"mixture_cons_{i}_{j}",
            )
    for i in range(n_wines_to_produce):
        scip.addCons(
            quicksum(wine_mixture_vars[i][k] for k in range(n_vineyards)) == 1,
            name=f"wine_mix_{i}",
        )
    for k in range(n_vineyards):
        scip.addCons(
            quicksum(wine_mixture_vars[i][k] for i in range(n_wines_to_produce))
            <= vineyard_litre_limits[k],
            name=f"vineyard_limit_{k}",
        )

    # Add the budget constraint
    scip.addCons(
        quicksum(
            quicksum(wine_mixture_vars[i][k] * vineyard_costs[k] for k in range(n_vineyards))
            for i in range(n_wines_to_produce)
        )
        <= budget,
        name=f"budget_cons",
    )

    # Add the ML constraint. Add in a single batch!
    pred_cons = add_predictor_constr(
        scip,
        reg,
        feature_vars,
        quality_vars,
        unique_naming_prefix="predictor_",
        epsilon=epsilon,
        formulation=formulation,
    )

    # Add a constraint ensuring minimum wine quality on those produced
    min_wine_quality = data_random_state.uniform(4.2, 4.5)
    for i in range(n_wines_to_produce):
        scip.addCons(quality_vars[i][0] >= min_wine_quality, name=f"min_quality_{i}")

    # Set the SCIP objective
    scip.setObjective(
        quicksum(-quality_vars[i][0] for i in range(n_wines_to_produce)) / n_wines_to_produce + 10
    )

    if not build_only:
        # Optimise the SCIP model
        scip.optimize()

        # We can check the "error" of the MIP embedding via the difference between SKLearn and SCIP output
        if np.max(pred_cons.get_error()) > 10**-3:
            error = np.max(pred_cons.get_error())
            raise AssertionError(f"Max error {error} exceeds threshold of {10 ** -3}")

    return scip


def test_wine_manufacturer_sk_rf():
    scip = build_and_optimise_wine_manufacturer(
        data_seed=42,
        training_seed=42,
        n_vineyards=15,
        n_wines_to_produce=2,
        framework="sklearn",
        gbdt_rf_or_mlp="rf",
        max_depth=3,
        n_estimators_layers=2,
    )


def test_wine_manufacturer_sk_gbdt():
    scip = build_and_optimise_wine_manufacturer(
        data_seed=42,
        training_seed=42,
        n_vineyards=10,
        n_wines_to_produce=2,
        framework="sklearn",
        gbdt_rf_or_mlp="gbdt",
        max_depth=3,
        n_estimators_layers=3,
    )


def test_wine_manufacturer_xgb_rf():
    scip = build_and_optimise_wine_manufacturer(
        data_seed=21,
        training_seed=21,
        n_vineyards=10,
        n_wines_to_produce=2,
        framework="xgboost",
        gbdt_rf_or_mlp="rf",
        max_depth=3,
        n_estimators_layers=3,
    )


def test_wine_manufacturer_xgb_gbdt():
    scip = build_and_optimise_wine_manufacturer(
        data_seed=21,
        training_seed=21,
        n_vineyards=10,
        n_wines_to_produce=2,
        framework="xgboost",
        gbdt_rf_or_mlp="gbdt",
        max_depth=3,
        n_estimators_layers=3,
    )


def test_wine_manufacturer_lgb_rf():
    scip = build_and_optimise_wine_manufacturer(
        data_seed=18,
        training_seed=18,
        n_vineyards=12,
        n_wines_to_produce=3,
        framework="lightgbm",
        gbdt_rf_or_mlp="rf",
        max_depth=3,
        n_estimators_layers=2,
    )


def test_wine_manufacturer_lgb_gbdt():
    scip = build_and_optimise_wine_manufacturer(
        data_seed=18,
        training_seed=18,
        n_vineyards=10,
        n_wines_to_produce=2,
        framework="lightgbm",
        gbdt_rf_or_mlp="gbdt",
        max_depth=3,
        n_estimators_layers=3,
    )


def test_wine_manufacturer_sklearn_mlp():
    scip = build_and_optimise_wine_manufacturer(
        data_seed=18,
        training_seed=18,
        n_vineyards=12,
        n_wines_to_produce=3,
        framework="sklearn",
        gbdt_rf_or_mlp="mlp",
        max_depth=3,
        n_estimators_layers=3,
        layer_size=7,
    )


def test_wine_manufacturer_sklearn_mlp_bigm():
    scip = build_and_optimise_wine_manufacturer(
        data_seed=18,
        training_seed=18,
        n_vineyards=15,
        n_wines_to_produce=3,
        framework="sklearn",
        gbdt_rf_or_mlp="mlp",
        formulation="bigm",
        max_depth=3,
        n_estimators_layers=3,
        layer_size=7,
    )
